"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.
"""

default_app_config = 'django_ledger.apps.DjangoLedgerConfig'

"""Django Ledger"""
__version__ = '0.8.0'
__license__ = 'GPLv3 License'

__author__ = 'Miguel Sanda'
__email__ = 'msanda@arrobalytics.com'

__url__ = 'https://www.djangoledger.com'



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
Miguel Sanda <msanda@arrobalytics.com>
"""

from django.forms import (ModelForm, TextInput, BooleanField, ValidationError, IntegerField,
                          EmailInput, URLInput, CheckboxInput, Select, Form)
from django.utils.translation import gettext_lazy as _

from django_ledger.forms.utils import validate_cszc
from django_ledger.models.entity import EntityModel
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES
from django.core.validators import MinValueValidator, MaxValueValidator


class EntityModelCreateForm(ModelForm):
    default_coa = BooleanField(required=False, initial=False, label=_('Populate Default CoA'))
    activate_all_accounts = BooleanField(required=False, initial=False, label=_('Activate All Accounts'))
    generate_sample_data = BooleanField(required=False, initial=False, label=_('Fill With Sample Data?'))

    def clean_name(self):
        name = self.cleaned_data.get('name')
        if not name:
            raise ValidationError(_('Please provide a valid name for new Entity.'))
        if len(name) < 3:
            raise ValidationError(_('Looks like this entity name is too short...'))
        return name

    def clean(self):
        populate_coa = self.cleaned_data['default_coa']
        activate_all_accounts = self.cleaned_data['activate_all_accounts']
        sample_data = self.cleaned_data['generate_sample_data']

        if sample_data and not all([
            populate_coa,
            activate_all_accounts
        ]):
            raise ValidationError(f'Filling sample data requires using default CoA and activate all accounts.')
        validate_cszc(self.cleaned_data)

    class Meta:
        model = EntityModel
        fields = [
            'name',
            'address_1',
            'address_2',
            'city',
            'state',
            'zip_code',
            'country',
            'email',
            'website',
            'phone',
            'fy_start_month',
            'activate_all_accounts',
            'accrual_method'
        ]
        labels = {
            'name': _('Entity Name'),
        }
        widgets = {
            'name': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-large',
                    'placeholder': _('Entity name...'),
                    'required': True
                }
            ),
            'address_1': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Address line 1')
            }),
            'address_2': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Address line 2')
            }),
            'city': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('City')
            }),
            'state': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('State')
            }),
            'zip_code': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Zip Code')
            }),
            'country': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Country')
            }),
            'phone': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Phone number...')
            }),
            'email': EmailInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Entity email...')
            }),
            'website': URLInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('http://www.mywebsite.com...')
            }),
            'default_coa': CheckboxInput(attrs={
                'class': 'checkbox'
            }),
            'fy_start_month': Select(attrs={
                'class': 'input'
            }),
            'accrual_method': CheckboxInput(attrs={
                'class': 'checkbook'
            })
        }


class EntityModelUpdateForm(ModelForm):
    class Meta:
        model = EntityModel
        fields = [
            'name',
            'address_1',
            'address_2',
            'city',
            'state',
            'zip_code',
            'country',
            'email',
            'phone',
            'website',
            'fy_start_month'
        ]
        labels = {
            'name': _('Entity Name')
        }
        widgets = {
            'name': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'placeholder': _('Entity name...')
                }
            ),
            'address_1': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'placeholder': _('Address line 1')
                }),
            'address_2': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'placeholder': _('Address line 2')
                }),
            'city': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'placeholder': _('City')
                }),
            'state': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'placeholder': _('State')
                }),
            'zip_code': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'placeholder': _('Zip Code')
                }),
            'country': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'placeholder': _('Country')
                }),
            'email': EmailInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'placeholder': _('Email...')
                }
            ),
            'phone': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'placeholder': _('Phone...')
                }
            ),
            'website': URLInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'placeholder': _('Website...')
                }
            ),
            'fy_start_month': Select(
                attrs={
                    'class': 'input'
                })
        }



from django.urls import path

from django_ledger.views import (
    ProductListView, ProductCreateView, UnitOfMeasureModelCreateView,
    UnitOfMeasureModelListView, UnitOfMeasureModelUpdateView, ProductUpdateView,
    ExpenseItemCreateView, ExpenseItemModelListView, ExpenseItemUpdateView, InventoryItemModelListView,
    InventoryItemCreateView, InventoryItemUpdateView, UnitOfMeasureModelDeleteView, ProductDeleteView,
    ServiceListView, ServiceCreateView, ServiceUpdateView, ServiceDeleteView
)

urlpatterns = [

    # Unit of Measure Views...
    path('<str:entity_slug>/uom/list/', UnitOfMeasureModelListView.as_view(), name='uom-list'),
    path('<str:entity_slug>/uom/create/', UnitOfMeasureModelCreateView.as_view(), name='uom-create'),
    path('<str:entity_slug>/uom/update/<uuid:uom_pk>/', UnitOfMeasureModelUpdateView.as_view(), name='uom-update'),
    path('<str:entity_slug>/uom/delete/<uuid:uom_pk>/', UnitOfMeasureModelDeleteView.as_view(), name='uom-delete'),

    # Product Views....
    path('<str:entity_slug>/product/list/', ProductListView.as_view(), name='product-list'),
    path('<str:entity_slug>/product/create/', ProductCreateView.as_view(), name='product-create'),
    path('<str:entity_slug>/product/update/<uuid:item_pk>/', ProductUpdateView.as_view(), name='product-update'),
    path('<str:entity_slug>/product/delete/<uuid:item_pk>/', ProductDeleteView.as_view(), name='product-delete'),

    # Service Views...
    path('<str:entity_slug>/service/list', ServiceListView.as_view(), name='service-list'),
    path('<str:entity_slug>/service/create/', ServiceCreateView.as_view(), name='service-create'),
    path('<str:entity_slug>/service/update/<uuid:item_pk>/', ServiceUpdateView.as_view(), name='service-update'),
    path('<str:entity_slug>/service/delete/<uuid:item_pk>/', ServiceDeleteView.as_view(), name='service-delete'),

    # Expense Views...
    path('<str:entity_slug>/expense/list/', ExpenseItemModelListView.as_view(), name='expense-list'),
    path('<str:entity_slug>/expense/create/', ExpenseItemCreateView.as_view(), name='expense-create'),
    path('<str:entity_slug>/expense/update/<uuid:item_pk>/', ExpenseItemUpdateView.as_view(), name='expense-update'),

    # Inventory Views...
    path('<str:entity_slug>/list/inventory/', InventoryItemModelListView.as_view(), name='inventory-item-list'),
    path('<str:entity_slug>/create/inventory/', InventoryItemCreateView.as_view(), name='inventory-item-create'),
    path('<str:entity_slug>/update/inventory/<uuid:item_pk>/', InventoryItemUpdateView.as_view(),
         name='inventory-item-update'),

]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
Miguel Sanda <msanda@arrobalytics.com>
"""

from django.core.exceptions import ValidationError
from django.forms import (ModelForm, DateInput, TextInput, Select, CheckboxInput, modelformset_factory, Textarea,
                          DateTimeInput)
from django.forms.models import BaseModelFormSet
from django.utils.translation import gettext_lazy as _

from django_ledger.io.roles import ASSET_CA_CASH, ASSET_CA_RECEIVABLES, LIABILITY_CL_DEFERRED_REVENUE
from django_ledger.models import (AccountModel, CustomerModel, InvoiceModel, ItemTransactionModel, ItemModel)
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class InvoiceModelCreateForEstimateForm(ModelForm):

    def __init__(self, *args, entity_slug, user_model, **kwargs):
        super().__init__(*args, **kwargs)
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model
        self.get_customer_queryset()
        self.get_accounts_queryset()

    def get_customer_queryset(self):
        if 'customer' in self.fields:
            customer_qs = CustomerModel.objects.for_entity(
                entity_model=self.ENTITY_SLUG
            ).active()
            self.fields['customer'].queryset = customer_qs

    def get_accounts_queryset(self):

        if all([
            'cash_account' in self.fields,
            'prepaid_account' in self.fields,
            'unearned_account' in self.fields,
        ]):

            account_qs = AccountModel.objects.for_entity(
                entity_model=self.ENTITY_SLUG
            ).for_invoice()

            self.fields['cash_account'].queryset = account_qs.filter(role__exact=ASSET_CA_CASH)
            self.fields['prepaid_account'].queryset = account_qs.filter(role__exact=ASSET_CA_RECEIVABLES)
            self.fields['unearned_account'].queryset = account_qs.filter(role__exact=LIABILITY_CL_DEFERRED_REVENUE)

    class Meta:
        model = InvoiceModel
        fields = [
            'terms',
            'cash_account',
            'prepaid_account',
            'unearned_account'
        ]
        labels = {
            'terms': _('Invoice Terms'),
            'date_draft': _('Draft Date'),
            'unearned_account': _('Deferred Revenue Account'),
            'prepaid_account': _('Receivable Account')
        }
        widgets = {
            'customer': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'date_draft': DateInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Invoice Date (YYYY-MM-DD)...'),
                'id': 'djl-invoice-draft-date-input'
            }),
            'amount_due': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': '$$$'}),
            'terms': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'cash_account': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'prepaid_account': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'unearned_account': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
        }


class InvoiceModelCreateForm(InvoiceModelCreateForEstimateForm):
    class Meta(InvoiceModelCreateForEstimateForm.Meta):
        fields = [
            'customer',
            'date_draft',
            'terms',
            'cash_account',
            'prepaid_account',
            'unearned_account'
        ]


class BaseInvoiceModelUpdateForm(ModelForm):

    def __init__(self,
                 *args,
                 entity_slug,
                 user_model,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model
        self.INVOICE_MODEL: InvoiceModel = self.instance

    class Meta:
        model = InvoiceModel
        fields = [
            'markdown_notes'
        ]
        labels = {
            'progress': _('Progress Amount 0.00 -> 1.00 (percent)'),
            'accrue': _('Will this Bill be Accrued?'),
            'amount_paid': _('Amount Received')
        }
        widgets = {
            'timestamp': DateTimeInput(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'terms': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'invoice_status': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'customer': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'date_paid': DateInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                    'placeholder': _('Paid Date (YYYY-MM-DD)...')}
            ),
            'amount_paid': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
                }),
            'progress': TextInput(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'accrue': CheckboxInput(attrs={'type': 'checkbox'}),
            'paid': CheckboxInput(attrs={'type': 'checkbox'}),
            'markdown_notes': Textarea(attrs={
                'class': 'textarea'
            })

        }


class DraftInvoiceModelUpdateForm(BaseInvoiceModelUpdateForm):
    class Meta(BaseInvoiceModelUpdateForm.Meta):
        fields = [
            'customer',
            'terms',
            'markdown_notes'
        ]


class InReviewInvoiceModelUpdateForm(BaseInvoiceModelUpdateForm):
    class Meta(BaseInvoiceModelUpdateForm.Meta):
        fields = [
            'markdown_notes'
        ]


class ApprovedInvoiceModelUpdateForm(BaseInvoiceModelUpdateForm):
    class Meta(BaseInvoiceModelUpdateForm.Meta):
        fields = [
            'amount_paid',
            'markdown_notes'
        ]


class AccruedAndApprovedInvoiceModelUpdateForm(BaseInvoiceModelUpdateForm):
    class Meta(BaseInvoiceModelUpdateForm.Meta):
        fields = [
            'progress',
            'amount_paid',
            'markdown_notes'
        ]


class PaidInvoiceModelUpdateForm(BaseInvoiceModelUpdateForm):
    class Meta(BaseInvoiceModelUpdateForm.Meta):
        fields = [
            'markdown_notes'
        ]


class InvoiceItemForm(ModelForm):

    def __init__(self,
                 *args,
                 entity_slug,
                 user_model,
                 invoice_model,
                 **kwargs):
        super(InvoiceItemForm, self).__init__(*args, **kwargs)
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model
        self.INVOICE_MODEL = invoice_model

    def clean(self):
        cleaned_data = super(InvoiceItemForm, self).clean()
        quantity = cleaned_data['quantity']
        if self.instance.item_model_id:
            item_model: ItemModel = self.instance.item_model
            if item_model.for_inventory and quantity > item_model.inventory_received:
                raise ValidationError(f'Cannot invoice more than {item_model.inventory_received} units available.')
        return cleaned_data

    class Meta:
        model = ItemTransactionModel
        fields = [
            'item_model',
            'unit_cost',
            'quantity'
        ]
        widgets = {
            'item_model': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'unit_cost': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'quantity': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            })
        }


class BaseInvoiceItemTransactionFormset(BaseModelFormSet):

    def __init__(self, *args,
                 entity_slug,
                 user_model,
                 invoice_model,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.USER_MODEL = user_model
        self.INVOICE_MODEL: InvoiceModel = invoice_model
        self.ENTITY_SLUG = entity_slug

        items_qs = ItemModel.objects.for_invoice(
            entity_model=self.ENTITY_SLUG
        )

        for form in self.forms:
            if not self.INVOICE_MODEL.can_edit_items():
                form.fields['item_model'].disabled = True
                form.fields['quantity'].disabled = True
                form.fields['unit_cost'].disabled = True
                form.can_delete = False
            form.fields['item_model'].queryset = items_qs

    def get_queryset(self):
        if not self.queryset:
            self.queryset = ItemTransactionModel.objects.for_invoice(
                entity_model=self.ENTITY_SLUG,
                invoice_pk=self.INVOICE_MODEL.uuid
            )
        else:
            self.queryset = self.INVOICE_MODEL.itemtransactionmodel_set.all()
        return self.queryset

    def get_form_kwargs(self, index):
        return {
            'entity_slug': self.ENTITY_SLUG,
            'user_model': self.USER_MODEL,
            'invoice_model': self.INVOICE_MODEL,
        }


def get_invoice_itemtxs_formset_class(invoice_model: InvoiceModel):
    can_delete = invoice_model.can_edit_items()
    return modelformset_factory(
        model=ItemTransactionModel,
        form=InvoiceItemForm,
        formset=BaseInvoiceItemTransactionFormset,
        can_delete=can_delete,
        extra=5 if can_delete else 0
    )



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
Miguel Sanda <msanda@arrobalytics.com>
"""
from django.urls import path, include

from django_ledger import views

app_name = 'django_ledger'

urlpatterns = [

    path('entity/', include('django_ledger.urls.entity')),
    path('unit/', include('django_ledger.urls.unit')),
    path('financial-statement/', include('django_ledger.urls.financial_statement')),
    path('chart-of-accounts/', include('django_ledger.urls.chart_of_accounts')),
    path('account/', include('django_ledger.urls.account')),
    path('ledger/', include('django_ledger.urls.ledger')),
    path('journal-entry/', include('django_ledger.urls.journal_entry')),
    path('transactions/', include('django_ledger.urls.transactions')),
    path('invoice/', include('django_ledger.urls.invoice')),
    path('bill/', include('django_ledger.urls.bill')),
    path('purchase_order/', include('django_ledger.urls.purchase_order')),
    path('customer/', include('django_ledger.urls.customer')),
    path('customer-job/', include('django_ledger.urls.estimate')),
    path('vendor/', include('django_ledger.urls.vendor')),
    path('item/', include('django_ledger.urls.item')),
    path('bank-account/', include('django_ledger.urls.bank_account')),
    path('data-import/', include('django_ledger.urls.data_import')),
    path('closing_entry/', include('django_ledger.urls.closing_entry')),
    path('auth/', include('django_ledger.urls.auth')),
    path('feedback/', include('django_ledger.urls.feedback')),
    path('inventory/', include('django_ledger.urls.inventory')),
    path('home/', include('django_ledger.urls.home')),
    path('api/', include('django_ledger.urls.djl_api')),
    path('', views.RootUrlView.as_view(), name='root'),
]




"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

This module implements the InvoiceModel, which represents the Sales Invoice/ Sales Invoice/ Tax Invoice/ Proof of Sale
which the :func:`EntityModel <django_ledger.models.entity.EntityModel>` issues to its customers for the supply of
goods or services. The model manages all the Sales Invoices which are issued by the :func:`EntityModel
<django_ledger.models.entity.EntityModel>`. In addition to tracking the invoice amount , it tracks the receipt and
due amount.

Examples
________
>>> user_model = request.user  # django UserModel
>>> entity_slug = kwargs['entity_slug'] # may come from view kwargs
>>> invoice_model = InvoiceModel()
>>> ledger_model, invoice_model = invoice_model.configure(entity_slug=entity_slug, user_model=user_model)
>>> invoice_model.save()
"""
import warnings
from datetime import date, datetime
from decimal import Decimal
from typing import Union, Optional, Tuple, Dict
from uuid import uuid4, UUID

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.db import models, transaction, IntegrityError
from django.db.models import Q, Sum, F, Count
from django.db.models.signals import pre_save
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from django_ledger.io import ASSET_CA_CASH, ASSET_CA_RECEIVABLES, LIABILITY_CL_DEFERRED_REVENUE
from django_ledger.io.io_core import get_localtime, get_localdate
from django_ledger.models import (
    lazy_loader, ItemTransactionModelQuerySet,
    ItemModelQuerySet, ItemModel, QuerySet, Manager
)
from django_ledger.models.deprecations import deprecated_entity_slug_behavior
from django_ledger.models.entity import EntityModel
from django_ledger.models.mixins import (
    CreateUpdateMixIn, AccrualMixIn,
    MarkdownNotesMixIn, PaymentTermsMixIn,
    ItemizeMixIn
)
from django_ledger.models.signals import (
    invoice_status_draft,
    invoice_status_in_review,
    invoice_status_approved,
    invoice_status_paid,
    invoice_status_canceled,
    invoice_status_void
)
from django_ledger.settings import DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING, DJANGO_LEDGER_INVOICE_NUMBER_PREFIX, \
    DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR

UserModel = get_user_model()


class InvoiceModelValidationError(ValidationError):
    pass


class InvoiceModelQuerySet(QuerySet):
    """
   A custom defined QuerySet for the InvoiceModel.
   This implements multiple methods or queries that we need to run to get a status of Invoices raised by the entity.
   For example, We might want to have list of invoices which are paid, unpaid, due , overDue, approved or in draft stage.
   All these separate functions will assist in making such queries and building customized reports.
   """

    def draft(self):
        """
        Default status of any invoice that is created.
        Draft invoices do not impact the Ledger.

        Returns
        -------
        InvoiceModelQuerySet
            Returns a QuerySet of draft invoices only.
        """
        return self.filter(invoice_status__exact=InvoiceModel.INVOICE_STATUS_DRAFT)

    def in_review(self):
        """
        In review invoices are those that need additional review or approvals before being approved.
        Draft invoices do not impact the Ledger.

        Returns
        -------
        InvoiceModelQuerySet
            Returns a QuerySet of in review invoices only.
        """
        return self.filter(invoice_status__exact=InvoiceModel.INVOICE_STATUS_DRAFT)

    def approved(self):
        """
        Approved invoices are those that have been reviewed and are expected to be paid before the due date.

        Returns
        -------
        InvoiceModelQuerySet
            Returns a QuerySet of approved invoices only.
        """
        return self.filter(invoice_status__exact=InvoiceModel.INVOICE_STATUS_APPROVED)

    def paid(self):
        """
        Paid invoices are those that have received 100% of the amount due.

        Returns
        -------
        InvoiceModelQuerySet
            Returns a QuerySet of paid invoices only.
        """
        return self.filter(invoice_status__exact=InvoiceModel.INVOICE_STATUS_PAID)

    def void(self):
        """
        Void invoices are those that where rolled back after being approved.
        Void invoices rollback all transactions by creating a new set of transactions posted on the date_void.

        Returns
        -------
        InvoiceModelQuerySet
            Returns a QuerySet of void invoices only.
        """
        return self.filter(invoice_status__exact=InvoiceModel.INVOICE_STATUS_VOID)

    def canceled(self):
        """
        Canceled invoices are those that are discarded during the draft or in review status.
        These invoices never had an impact on the books.

        Returns
        -------
        InvoiceModelQuerySet
            Returns a QuerySet of canceled invoices only.
        """
        return self.filter(invoice_status__exact=InvoiceModel.INVOICE_STATUS_CANCELED)

    def active(self):
        """
        Active invoices are those that are approved or paid, which have
        impacted or have the potential to impact the Entity's Ledgers.

        Returns
        -------
        InvoiceModelQuerySet
            Returns a QuerySet of active invoices only.
        """
        return self.filter(
            Q(invoice_status__exact=InvoiceModel.INVOICE_STATUS_APPROVED) |
            Q(invoice_status__exact=InvoiceModel.INVOICE_STATUS_PAID)
        )

    def overdue(self):
        """
        Overdue invoices are those which due date is in the past.

        Returns
        -------
        InvoiceModelQuerySet
            Returns a QuerySet of overdue invoices only.
        """
        return self.filter(date_due__lt=get_localdate())

    def unpaid(self):
        """
        Unpaid invoices are those that are approved but have not received 100% of the amount due.
        Equivalent to approved().

        Returns
        -------
        InvoiceModelQuerySet
            Returns a QuerySet of paid invoices only.
        """
        return self.filter(invoice_status__exact=InvoiceModel.INVOICE_STATUS_APPROVED)

    def for_user(self, user_model):
        if user_model.is_superuser:
            return self
        return self.filter(
            Q(ledger__entity__admin=user_model) |
            Q(ledger__entity__managers__in=[user_model])
        )


class InvoiceModelManager(Manager):
    """
    A custom defined InvoiceModel Manager that will act as an interface to handling the DB queries to the InvoiceModel.
    The default "get_queryset" has been overridden to refer the custom defined "InvoiceModelQuerySet"
    """

    def get_queryset(self) -> InvoiceModelQuerySet:
        qs = InvoiceModelQuerySet(self.model, using=self._db)
        return qs.select_related(
            'ledger',
            'ledger__entity'
        )

    @deprecated_entity_slug_behavior
    def for_entity(self, entity_model: EntityModel | str | UUID = None, **kwargs) -> InvoiceModelQuerySet:
        """
        Returns a QuerySet of InvoiceModels associated with a specific EntityModel & UserModel.
        May pass an instance of EntityModel or a String representing the EntityModel slug.

        Parameters
        ----------
        entity_model: str or EntityModel
            The entity slug or EntityModel used for filtering the QuerySet.

        Returns
        -------
        InvoiceModelQuerySet
            A Filtered InvoiceModelQuerySet.
        """
        qs = self.get_queryset()

        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(ledger__entity=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(ledger__entity_id=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(ledger__entity__slug__exact=entity_model)
        else:
            raise InvoiceModelValidationError(
                message='Must provide either a string, UUID or an EntityModel',
            )
        return qs


class InvoiceModelAbstract(
    AccrualMixIn,
    ItemizeMixIn,
    PaymentTermsMixIn,
    MarkdownNotesMixIn,
    CreateUpdateMixIn
):
    """
    This is the main abstract class which the InvoiceModel database will inherit from.
    The InvoiceModel inherits functionality from the following MixIns:
    
        1. :func:`LedgerWrapperMixIn <django_ledger.models.mixins.LedgerWrapperMixIn>`
        2. :func:`PaymentTermsMixIn <django_ledger.models.mixins.PaymentTermsMixIn>`
        3. :func:`MarkdownNotesMixIn <django_ledger.models.mixins.MarkdownNotesMixIn>`
        4. :func:`CreateUpdateMixIn <django_ledger.models.mixins.CreateUpdateMixIn>`


    Attributes
    ----------
    uuid: UUID
        This is a unique primary key generated for the table. The default value of this field is uuid4().
    invoice_number: str
        Auto assigned number at creation by generate_invoice_number() function.
        Prefix be customized with DJANGO_LEDGER_INVOICE_NUMBER_PREFIX setting.
        Includes a reference to the Fiscal Year, Entity Unit and a sequence number. Max Length is 20.
    invoice_status: str
        Current status of the InvoiceModel. Must be one of the choices as mentioned under "INVOICE_STATUS".
        By default, the status will be "Draft".
    customer: :obj:`CustomerModel`
        This is the foreign key reference to the CustomerModel from whom the purchase has been made.
    additional_info: dict
        Any additional metadata about the InvoiceModel may be stored here as a dictionary object.
        The data is serialized and stored as a JSON document in the Database.
    invoice_items:
        A foreign key reference to the list of ItemTransactionModel that make the invoice amount.
    ce_model: EstimateModel
        A foreign key to the InvoiceModel associated EstimateModel for overall Job/Contract tracking.
    date_draft: date
        The draft date represents the date when the InvoiceModel was first created.
        Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    date_in_review: date
        The in review date represents the date when the InvoiceModel was marked as In Review status.
        Will be null if InvoiceModel is canceled during draft status. Defaults to
        :func:`localdate <django.utils.timezone.localdate>`.
    date_approved: date
        The approved date represents the date when the InvoiceModel was approved.
        Will be null if InvoiceModel is canceled.
        Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    date_paid: date
        The paid date represents the date when the InvoiceModel was paid and amount_due equals amount_paid.
        Will be null if InvoiceModel is canceled. Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    date_void: date
        The void date represents the date when the InvoiceModel was void, if applicable.
        Will be null unless InvoiceModel is void. Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    date_canceled: date
        The canceled date represents the date when the InvoiceModel was canceled, if applicable.
        Will be null unless InvoiceModel is canceled. Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    """

    IS_DEBIT_BALANCE = True
    REL_NAME_PREFIX = 'invoice'

    INVOICE_STATUS_DRAFT = 'draft'
    INVOICE_STATUS_REVIEW = 'in_review'
    INVOICE_STATUS_APPROVED = 'approved'
    INVOICE_STATUS_PAID = 'paid'
    INVOICE_STATUS_VOID = 'void'
    INVOICE_STATUS_CANCELED = 'canceled'

    INVOICE_STATUS = [
        (INVOICE_STATUS_DRAFT, _('Draft')),
        (INVOICE_STATUS_REVIEW, _('In Review')),
        (INVOICE_STATUS_APPROVED, _('Approved')),
        (INVOICE_STATUS_PAID, _('Paid')),
        (INVOICE_STATUS_VOID, _('Void')),
        (INVOICE_STATUS_CANCELED, _('Canceled'))
    ]
    """
    The different invoice status options and their representation in the Database.
    """

    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    entity_model = models.ForeignKey('django_ledger.EntityModel',
                                     on_delete=models.CASCADE,
                                     null=True,
                                     blank=True,
                                     editable=False)
    invoice_number = models.SlugField(max_length=20,
                                      editable=False,
                                      verbose_name=_('Invoice Number'))
    invoice_status = models.CharField(max_length=10, choices=INVOICE_STATUS, default=INVOICE_STATUS[0][0],
                                      verbose_name=_('Invoice Status'))
    customer = models.ForeignKey('django_ledger.CustomerModel',
                                 on_delete=models.RESTRICT,
                                 verbose_name=_('Customer'))

    cash_account = models.ForeignKey('django_ledger.AccountModel',
                                     on_delete=models.RESTRICT,
                                     null=True,
                                     blank=True,
                                     verbose_name=_('Cash Account'),
                                     related_name=f'{REL_NAME_PREFIX}_cash_account')
    prepaid_account = models.ForeignKey('django_ledger.AccountModel',
                                        on_delete=models.RESTRICT,
                                        null=True,
                                        blank=True,
                                        verbose_name=_('Prepaid Account'),
                                        related_name=f'{REL_NAME_PREFIX}_prepaid_account')
    unearned_account = models.ForeignKey('django_ledger.AccountModel',
                                         on_delete=models.RESTRICT,
                                         null=True,
                                         blank=True,
                                         verbose_name=_('Unearned Account'),
                                         related_name=f'{REL_NAME_PREFIX}_unearned_account')

    additional_info = models.JSONField(blank=True,
                                       null=True,
                                       default=dict,
                                       verbose_name=_('Invoice Additional Info'))
    invoice_items = models.ManyToManyField('django_ledger.ItemModel',
                                           through='django_ledger.ItemTransactionModel',
                                           through_fields=('invoice_model', 'item_model'),
                                           verbose_name=_('Invoice Items'))

    ce_model = models.ForeignKey('django_ledger.EstimateModel',
                                 on_delete=models.RESTRICT,
                                 null=True,
                                 blank=True,
                                 verbose_name=_('Associated Customer Job/Estimate'))

    date_draft = models.DateField(null=True, blank=True, verbose_name=_('Draft Date'))
    date_in_review = models.DateField(null=True, blank=True, verbose_name=_('In Review Date'))
    date_approved = models.DateField(null=True, blank=True, verbose_name=_('Approved Date'))
    date_paid = models.DateField(null=True, blank=True, verbose_name=_('Paid Date'))
    date_void = models.DateField(null=True, blank=True, verbose_name=_('Void Date'))
    date_canceled = models.DateField(null=True, blank=True, verbose_name=_('Canceled Date'))

    objects = InvoiceModelManager.from_queryset(queryset_class=InvoiceModelQuerySet)()

    class Meta:
        abstract = True
        ordering = ['-updated']
        verbose_name = _('Invoice')
        verbose_name_plural = _('Invoices')
        indexes = [
            models.Index(fields=['invoice_status']),
            models.Index(fields=['terms']),

            models.Index(fields=['cash_account']),
            models.Index(fields=['prepaid_account']),
            models.Index(fields=['unearned_account']),

            models.Index(fields=['date_due']),
            models.Index(fields=['date_draft']),
            models.Index(fields=['date_in_review']),
            models.Index(fields=['date_approved']),
            models.Index(fields=['date_paid']),
            models.Index(fields=['date_canceled']),
            models.Index(fields=['date_void']),

            models.Index(fields=['customer']),
            models.Index(fields=['invoice_number']),
        ]

    def __str__(self):
        return f'Invoice: {self.invoice_number} | {self.get_invoice_status_display()}'

    def is_configured(self) -> bool:
        return all([
            super().is_configured(),
            self.invoice_status
        ])

    def configure(self,
                  entity_slug: Union[EntityModel, str],
                  user_model: Optional[UserModel] = None,
                  date_draft: Optional[date] = None,
                  ledger_posted: bool = False,
                  ledger_name: Optional[str] = None,
                  commit: bool = False,
                  commit_ledger: bool = False):
        """
        A configuration hook which executes all initial InvoiceModel setup on to the LedgerModel and all initial
        values of the InvoiceModel. Can only call this method once in the lifetime of a InvoiceModel.

        Parameters
        __________

        entity_slug: str or EntityModel
            The entity slug or EntityModel to associate the Invoice with.
        user_model: UserModel
            The UserModel making the request to check for QuerySet permissions.
        ledger_posted: bool
            An option to mark the InvoiceModel Ledger as posted at the time of configuration. Defaults to False.
        ledger_name: str
            Optional additional InvoiceModel ledger name or description.
        invoice_desc: str
            An optional description appended to the LedgerModel name.
        commit: bool
            Saves the current InvoiceModel after being configured.
        commit_ledger: bool
            Saves the InvoiceModel's LedgerModel while being configured.
        Returns
        -------
        A tuple of LedgerModel, InvoiceModel
        """

        if not self.is_configured():
            if isinstance(entity_slug, str):
                if not user_model:
                    raise InvoiceModelValidationError(_('Must pass user_model when using entity_slug.'))
                entity_qs = EntityModel.objects.for_user(user_model=user_model)
                entity_model: EntityModel = get_object_or_404(entity_qs, slug__exact=entity_slug)
            elif isinstance(entity_slug, EntityModel):
                entity_model = entity_slug
            else:
                raise InvoiceModelValidationError('entity_slug must be an instance of str or EntityModel')

            if entity_model.is_accrual_method():
                self.accrue = True
                self.progress = Decimal.from_float(1.00)
            else:
                self.accrue = False

            self.invoice_status = self.INVOICE_STATUS_DRAFT
            self.date_draft = get_localdate() if not date_draft else date_draft

            LedgerModel = lazy_loader.get_ledger_model()
            ledger_model: LedgerModel = LedgerModel(entity=entity_model, posted=ledger_posted)
            ledger_model.configure_for_wrapper_model(model_instance=self)
            ledger_name = f'Invoice {self.uuid}' if not ledger_name else ledger_name
            ledger_model.name = ledger_name
            ledger_model.clean()
            ledger_model.clean_fields()
            self.ledger = ledger_model

            if commit_ledger or commit:
                self.ledger.save()

            if self.can_generate_invoice_number():
                self.generate_invoice_number(commit=commit)
                ledger_model.ledger_xid = f'invoice-{self.invoice_number.lower()}-{str(ledger_model.entity_id)[-5:]}'
                ledger_model.save(update_fields=['ledger_xid'])

            self.clean()

            if commit:
                self.save()

        return self.ledger, self

    # ### ItemizeMixIn implementation START...

    def can_migrate_itemtxs(self) -> bool:
        return self.is_draft()

    def migrate_itemtxs(self, itemtxs: Dict, operation: str, commit: bool = False):
        itemtxs_batch = super().migrate_itemtxs(itemtxs=itemtxs, commit=commit, operation=operation)
        self.update_amount_due(itemtxs_qs=itemtxs_batch)
        self.get_state(commit=True)

        if commit:
            self.save(update_fields=['amount_due',
                                     'amount_receivable',
                                     'amount_unearned',
                                     'amount_earned',
                                     'updated'])
        return itemtxs_batch

    def get_item_model_qs(self) -> ItemModelQuerySet:
        return ItemModel.objects.filter(
            entity_id__exact=self.ledger.entity_id
        ).invoices()

    def validate_itemtxs_qs(self, queryset: ItemTransactionModelQuerySet):
        """
        Validates that the entire ItemTransactionModelQuerySet is bound to the InvoiceModel.

        Parameters
        ----------
        queryset: ItemTransactionModelQuerySet
            ItemTransactionModelQuerySet to validate.
        """
        valid = all([
            i.invoice_model_id == self.uuid for i in queryset
        ])
        if not valid:
            raise InvoiceModelValidationError(f'Invalid queryset. All items must be assigned to Invoice {self.uuid}')

    def get_itemtxs_data(self,
                         queryset: ItemTransactionModelQuerySet = None,
                         aggregate_on_db: bool = False,
                         lazy_agg: bool = False,
                         ) -> Tuple[ItemTransactionModelQuerySet, Dict]:
        """
        Fetches the InvoiceModel Items and aggregates the QuerySet.

        Parameters
        __________
        queryset:
            Optional pre-fetched ItemModelQueryset to use. Avoids additional DB query if provided.

        Returns
        _______
        A tuple: ItemTransactionModelQuerySet, dict
        """

        if not queryset:
            queryset = self.itemtransactionmodel_set.all().select_related(
                'item_model',
                'entity_unit',
                'po_model',
                'invoice_model'
            )
        else:
            self.validate_itemtxs_qs(queryset)

        if aggregate_on_db and isinstance(queryset, ItemTransactionModelQuerySet):
            return queryset, queryset.aggregate(
                total_amount__sum=Sum('total_amount'),
                total_items=Count('uuid')
            )

        return queryset, {
            'total_amount__sum': sum(i.total_amount for i in queryset),
            'total_items': len(queryset)
        } if not lazy_agg else None

    # ### ItemizeMixIn implementation END...

    def get_transaction_queryset(self, annotated: bool = False):
        """
        Fetches the TransactionModelQuerySet associated with the InvoiceModel instance.
        """
        TransactionModel = lazy_loader.get_txs_model()
        transaction_model_qs = TransactionModel.objects.all().for_ledger(ledger_model=self.ledger_id)
        if annotated:
            return transaction_model_qs.with_annotated_details()
        return transaction_model_qs

    def get_migrate_state_desc(self):
        """
        Description used when migrating transactions into the LedgerModel.

        Returns
        _______
        str
            Description as a string.
        """
        return f'Invoice {self.invoice_number} account adjustment.'

    def get_migration_data(self,
                           queryset: Optional[ItemTransactionModelQuerySet] = None) -> ItemTransactionModelQuerySet:

        """
        Fetches necessary item transaction data to perform a migration into the LedgerModel.

        Parameters
        ----------
        queryset: ItemTransactionModelQuerySet
            Optional pre-fetched ItemModelTransactionQueryset to use. Avoids additional DB query if provided.
        """
        if not queryset:
            queryset = self.itemtransactionmodel_set.all()
        else:
            self.validate_itemtxs_qs(queryset)

        return queryset.select_related('item_model').order_by(
            'item_model__earnings_account__uuid',
            'entity_unit__uuid',
            'item_model__earnings_account__balance_type').values(
            'item_model__earnings_account__uuid',
            'item_model__earnings_account__balance_type',
            'item_model__cogs_account__uuid',
            'item_model__cogs_account__balance_type',
            'item_model__inventory_account__uuid',
            'item_model__inventory_account__balance_type',
            'item_model__inventory_received',
            'item_model__inventory_received_value',
            'entity_unit__slug',
            'entity_unit__uuid',
            'quantity',
            'total_amount').annotate(
            account_unit_total=Sum('total_amount'))

    def update_amount_due(self,
                          itemtxs_qs: Optional[ItemTransactionModelQuerySet] = None) -> ItemTransactionModelQuerySet:
        """
        Updates the InvoiceModel amount due.

        Parameters
        ----------
        itemtxs_qs: ItemTransactionModelQuerySet
            Optional pre-fetched ItemTransactionModelQuerySet. Avoids additional DB if provided.
            Queryset is validated if provided.

        Returns
        -------
        ItemTransactionModelQuerySet
            Newly fetched of previously fetched ItemTransactionModelQuerySet if provided.
        """
        itemtxs_qs, itemtxs_agg = self.get_itemtxs_data(queryset=itemtxs_qs)
        self.amount_due = round(itemtxs_agg['total_amount__sum'], 2)
        return itemtxs_qs

    # STATE...
    def is_draft(self) -> bool:
        """
        Checks if the InvoiceModel is in Draft status.

        Returns
        _______
        bool
            True if InvoiceModel is Draft, else False.
        """
        return self.invoice_status == self.INVOICE_STATUS_DRAFT

    def is_review(self) -> bool:
        """
        Checks if the InvoiceModel is In Review status.

        Returns
        _______
        bool
            True if InvoiceModel is in Review, else False.
        """
        return self.invoice_status == self.INVOICE_STATUS_REVIEW

    def is_approved(self) -> bool:
        """
        Checks if the InvoiceModel is in Approved status.

        Returns
        _______
        bool
            True if InvoiceModel is Approved, else False.
        """
        return self.invoice_status == self.INVOICE_STATUS_APPROVED

    def is_paid(self) -> bool:
        """
        Checks if the InvoiceModel is in Paid status.

        Returns
        _______
        bool
            True if InvoiceModel is Paid, else False.
        """
        return self.invoice_status == self.INVOICE_STATUS_PAID

    def is_active(self):
        """
        Checks if the InvoiceModel has the potential to impact the books and produce financial statements status.

        Returns
        _______
        bool
            True if InvoiceModel is Active, else False.
        """
        return any([
            self.is_paid(),
            self.is_approved(),
            self.is_void()
        ])

    def is_canceled(self) -> bool:
        """
        Checks if the InvoiceModel is in Canceled status.

        Returns
        _______
        bool
            True if InvoiceModel is Canceled, else False.
        """
        return self.invoice_status == self.INVOICE_STATUS_CANCELED

    def is_void(self) -> bool:
        """
        Checks if the InvoiceModel is in Void status.

        Returns
        _______
        bool
            True if InvoiceModel is Void, else False.
        """
        return self.invoice_status == self.INVOICE_STATUS_VOID

    def is_past_due(self) -> bool:
        """
        Checks if the InvoiceModel is past due.

        Returns
        -------
        bool
            True if InvoiceModel is past due, else False.
        """
        if self.date_due and self.is_approved():
            return self.date_due < get_localdate()
        return False

    # PERMISSIONS....
    def can_draft(self):
        """
        Checks if the InvoiceModel can be marked as Draft.

        Returns
        -------
        bool
            True if InvoiceModel can be marked as draft, else False.
        """
        return self.is_review()

    def can_review(self):
        """
        Checks if the InvoiceModel can be marked as In Review.

        Returns
        -------
        bool
            True if InvoiceModel can be marked as in review, else False.
        """
        return all([
            self.is_configured(),
            self.is_draft()
        ])

    def can_approve(self):
        """
        Checks if the InvoiceModel can be marked as Approved.

        Returns
        -------
        bool
            True if InvoiceModel can be marked as approved, else False.
        """
        return self.is_review()

    def can_pay(self):
        """
        Checks if the InvoiceModel can be marked as Paid.

        Returns
        -------
        bool
            True if InvoiceModel can be marked as paid, else False.
        """
        return self.is_approved()

    def can_delete(self):
        """
        Checks if the InvoiceModel can be deleted.

        Returns
        -------
        bool
            True if InvoiceModel can be deleted, else False.
        """
        return any([
            self.is_review(),
            self.is_draft(),
            not self.ledger.is_locked()
        ])

    def can_void(self):
        """
        Checks if the InvoiceModel can be marked as Void status.

        Returns
        -------
        bool
            True if InvoiceModel can be marked as void, else False.
        """
        return all([
            self.is_approved(),
            float(self.amount_paid) == 0.00
        ])

    def can_cancel(self):
        """
        Checks if the InvoiceModel can be marked as Canceled status.

        Returns
        -------
        bool
            True if InvoiceModel can be marked as canceled, else False.
        """
        return any([
            self.is_draft(),
            self.is_review()
        ])

    def can_edit_items(self):
        """
        Checks if the InvoiceModel item transactions can be edited.

        Returns
        -------
        bool
            True if InvoiceModel items can be edited, else False.
        """
        return self.is_draft()

    def can_migrate(self):
        """
        Checks if the InvoiceModel can be migrated.

        Returns
        -------
        bool
            True if InvoiceModel can be migrated, else False.
        """
        can_migrate = super().can_migrate()
        if not can_migrate:
            return False
        return self.is_approved()

    def can_bind_estimate(self, estimate_model, raise_exception: bool = False) -> bool:
        """
        Checks if the InvoiceModel can be bound to a given EstimateModel.

        Parameters
        __________

        estimate_model: EstimateModel
            EstimateModel to check against.

        raise_exception: bool
            If True, raises InvoiceModelValidationError if unable to bind. Else, returns False.

        Returns
        _______

        bool
            True if can bind provided EstimateModel, else False.
        """

        if self.ce_model_id:
            if raise_exception:
                raise InvoiceModelValidationError(f'Invoice {self.invoice_number} already bound to '
                                                  f'Estimate {self.ce_model.estimate_number}')
            return False

        is_approved = estimate_model.is_approved()
        if not is_approved and raise_exception:
            raise InvoiceModelValidationError(f'Cannot bind estimate that is not approved.')
        return all([
            is_approved
        ])

    def can_generate_invoice_number(self):
        """
        Checks if InvoiceModel can generate its Document Number.

        Returns
        _______

        bool
            True if InvoiceModel can generate its invoice_number, else False.
        """
        return all([
            self.date_draft,
            not self.invoice_number,
            self.is_configured()
        ])

    # ACTIONS...

    def can_make_payment(self) -> bool:
        """
        Checks if the BillModel can accept a payment.


        Returns
        _______

        bool
            True if can bind provided PurchaseOderModel, else False.
        """
        return self.is_approved()

    def make_payment(self,
                     payment_amount: Union[Decimal, float, int],
                     payment_date: Optional[Union[datetime, date]] = None,
                     commit: bool = False,
                     raise_exception: bool = True):
        """
        Makes a payment to the InvoiceModel.

        Parameters
        __________

        payment_amount: Decimal ot float
            The payment amount to process.

        payment_date: datetime or date.
            Date or timestamp of the payment being applied.

        commit: bool
            If True, commits the transaction into the DB. Defaults to False.

        raise_exception: bool
            If True, raises InvoiceModelValidationError if payment exceeds amount due, else False.

        Returns
        _______

        bool
            True if can make payment, else False.
        """

        if isinstance(payment_amount, float):
            payment_amount = Decimal.from_float(payment_amount)
        elif isinstance(payment_amount, int):
            payment_amount = Decimal.from_float(float(payment_amount))
        self.amount_paid += payment_amount

        if self.amount_paid > self.amount_due:
            if raise_exception:
                raise InvoiceModelValidationError(
                    f'Amount paid: {self.amount_paid} exceed amount due: {self.amount_due}.'
                )
            return

        self.get_state(commit=True)
        self.clean()

        if not payment_date:
            payment_date = get_localtime()

        if commit:
            self.migrate_state(
                user_model=None,
                entity_slug=self.ledger.entity.slug,
                je_timestamp=payment_date,
                raise_exception=True
            )
            self.save(
                update_fields=[
                    'amount_paid',
                    'amount_earned',
                    'amount_unearned',
                    'amount_receivable',
                    'updated'
                ])

    def bind_estimate(self, estimate_model, commit: bool = False):
        """
        Binds InvoiceModel to a given EstimateModel. Raises ValueError if EstimateModel cannot be bound.

        Parameters
        __________
        estimate_model: EstimateModel
            EstimateModel to bind.

        raise_exception: bool
            Raises InvoiceModelValidationError if unable to bind EstimateModel.

        commit: bool
            Commits transaction into current InvoiceModel.
        """

        try:
            self.can_bind_estimate(estimate_model, raise_exception=True)
        except ValueError as e:
            raise e

        self.ce_model = estimate_model
        self.customer_id = estimate_model.customer_id
        self.clean()
        if commit:
            self.save(update_fields=[
                'ce_model',
                'customer_id',
                'updated'
            ])

    # DRAFT...
    def mark_as_draft(self, draft_date: Union[date, datetime], commit: bool = False, **kwargs):
        """
        Marks InvoiceModel as Draft.

        Parameters
        __________

        date_draft: date
            Draft date. If None, defaults to localdate().

        commit: bool
            Commits transaction into the Database. Defaults to False.
        """
        if not self.can_draft():
            raise InvoiceModelValidationError(f'Cannot mark PO {self.uuid} as draft...')

        if draft_date:
            if isinstance(draft_date, datetime):
                self.draft_date = draft_date.date()
            elif isinstance(draft_date, date):
                self.draft_date = draft_date
        else:
            self.draft_date = get_localdate()

        self.invoice_status = self.INVOICE_STATUS_DRAFT
        self.clean()
        if commit:
            self.save(update_fields=[
                'invoice_status',
                'updated'
            ])
        invoice_status_draft.send_robust(sender=self.__class__,
                                         instance=self,
                                         commited=commit, **kwargs)

    def get_mark_as_draft_html_id(self):
        """
        InvoiceModel Mark as Draft HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String
        """
        return f'djl-{self.uuid}-invoice-mark-as-draft'

    def get_mark_as_draft_url(self):
        """
        InvoiceModel Mark-as-Draft action URL.

        Parameters
        __________
        entity_slug: str
            Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
            is not cached on QuerySet.

        Returns
        _______
        str
            HTML ID as a String
        """
        return reverse('django_ledger:invoice-action-mark-as-draft',
                       kwargs={
                           'entity_slug': self.ledger.entity.slug,
                           'invoice_pk': self.uuid
                       })

    def get_mark_as_draft_message(self):
        """
        Internationalized confirmation message with Invoice Number.

        Returns
        _______
        str
            Mark-as-Draft InvoiceModel confirmation message as a String.
        """
        return _('Do you want to mark Invoice %s as Draft?') % self.invoice_number

    # REVIEW...
    def mark_as_review(self,
                       date_in_review: date = None,
                       itemtxs_qs=None,
                       commit: bool = False,
                       **kwargs):
        """
        Marks InvoiceModel as In Review.

        Parameters
        __________

        date_in_review: date
            InvoiceModel in review date. Defaults to localdate() if None.
        itemtxs_qs: ItemTransactionModelQuerySet
            Pre fetched ItemTransactionModelQuerySet to use. Avoids additional DB Query if previously fetched.
        commit: bool
            Commits transaction into the Database. Defaults to False.
        raise_exception: bool
            Raises InvoiceModelValidationError if InvoiceModel cannot be marked as in review. Defaults to True.
        """
        if not self.can_review():
            raise InvoiceModelValidationError(f'Cannot mark PO {self.uuid} as In Review...')

        self.date_in_review = get_localdate() if not date_in_review else date_in_review

        if not itemtxs_qs:
            itemtxs_qs = self.itemtransactionmodel_set.all()
        if not itemtxs_qs.count():
            raise InvoiceModelValidationError(message='Cannot review an Invoice without items...')
        if not self.amount_due:
            raise InvoiceModelValidationError(
                f'PO {self.invoice_number} cannot be marked as in review. Amount due must be greater than 0.'
            )

        self.invoice_status = self.INVOICE_STATUS_REVIEW
        self.clean()
        if commit:
            self.save(update_fields=[
                'invoice_status',
                'date_in_review',
                'updated'
            ])
        invoice_status_in_review.send_robust(sender=self.__class__,
                                             instance=self,
                                             commited=commit,
                                             **kwargs)

    def get_mark_as_review_html_id(self):
        """
        InvoiceModel Mark as In Review HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-{self.uuid}-invoice-mark-as-review'

    def get_mark_as_review_url(self):
        """
        InvoiceModel Mark-as-Review action URL.

        Parameters
        __________
        entity_slug: str
            Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
            is not cached on QuerySet.

        Returns
        _______
        str
            InvoiceModel mark-as-review action URL.
        """
        return reverse('django_ledger:invoice-action-mark-as-review',
                       kwargs={
                           'entity_slug': self.ledger.entity.slug,
                           'invoice_pk': self.uuid
                       })

    def get_mark_as_review_message(self):
        """
        Internationalized confirmation message with Invoice Number.

        Returns
        _______
        str
            Mark-as-Review InvoiceModel confirmation message as a String.
        """
        return _('Do you want to mark Invoice %s as In Review?') % self.invoice_number

    # APPROVED...
    def mark_as_approved(self,
                         entity_slug,
                         user_model,
                         date_approved: date = None,
                         commit: bool = False,
                         force_migrate: bool = False,
                         raise_exception: bool = True,
                         **kwargs):
        """
        Marks InvoiceModel as Approved.

        Parameters
        __________

        entity_slug
            Entity slug associated with the InvoiceModel. Avoids additional DB query if passed.

        user_model
            UserModel associated with request.

        date_approved: date
            InvoiceModel approved date. Defaults to localdate().

        commit: bool
            Commits transaction into the Database. Defaults to False.

        force_migrate: bool
            Forces migration. True if Accounting Method is Accrual.
        """
        if not self.can_approve():
            if raise_exception:
                raise InvoiceModelValidationError(f'Cannot mark PO {self.uuid} as Approved...')
            return

        self.invoice_status = self.INVOICE_STATUS_APPROVED

        if date_approved:
            if isinstance(date_approved, datetime):
                self.date_approved = date_approved.date()
            elif isinstance(date_approved, date):
                self.draft_date = date_approved
        else:
            self.date_approved = get_localdate()

        self.clean()
        if commit:
            self.save()
            if force_migrate or self.accrue:
                # normally no transactions will be present when marked as approved...
                self.migrate_state(
                    entity_slug=entity_slug,
                    user_model=user_model,
                    je_timestamp=date_approved,
                    force_migrate=self.accrue
                )
            self.ledger.post(commit=commit, raise_exception=raise_exception)
        invoice_status_approved.send_robust(sender=self.__class__,
                                            instance=self,
                                            commited=commit,
                                            **kwargs)

    def get_mark_as_approved_html_id(self):
        """
        InvoiceModel Mark as Approved HTML ID.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-{self.uuid}-invoice-mark-as-approved'

    def get_mark_as_approved_url(self):
        """
        InvoiceModel Mark-as-Approved action URL.

        Parameters
        __________
        entity_slug: str
            Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
            is not cached on QuerySet.

        Returns
        _______
        str
            InvoiceModel mark-as-approved action URL.
        """
        return reverse('django_ledger:invoice-action-mark-as-approved',
                       kwargs={
                           'entity_slug': self.ledger.entity.slug,
                           'invoice_pk': self.uuid
                       })

    def get_mark_as_approved_message(self):
        """
        Internationalized confirmation message with Invoice Number.

        Returns
        _______
        str
            Mark-as-Approved InvoiceModel confirmation message as a String.
        """
        return _('Do you want to mark Invoice %s as Approved?') % self.invoice_number

    # PAID...
    def mark_as_paid(self,
                     entity_slug: str,
                     user_model,
                     date_paid: date = None,
                     commit: bool = False,
                     **kwargs):
        """
        Marks InvoiceModel as Paid.

        Parameters
        __________

        entity_slug: str
            Entity slug associated with the InvoiceModel. Avoids additional DB query if passed.

        user_model:
            UserModel associated with request.

        date_paid: date
            InvoiceModel paid date. Defaults to localdate() if None.

        itemtxs_qs: ItemTransactionModelQuerySet
            Pre-fetched ItemTransactionModelQuerySet. Avoids additional DB query. Validated if passed.

        commit: bool
            Commits transaction into the Database. Defaults to False.
        """

        if not self.can_pay():
            raise InvoiceModelValidationError(f'Cannot mark PO {self.uuid} as Paid...')

        self.progress = Decimal.from_float(1.0)
        self.amount_paid = self.amount_due

        if date_paid:
            if isinstance(date_paid, datetime):
                self.date_paid = date_paid.date()
            elif isinstance(date_paid, date):
                self.date_paid = date_paid
        else:
            self.date_paid = get_localdate()

        if self.date_paid > get_localdate():
            raise InvoiceModelValidationError(f'Cannot pay {self.__class__.__name__} in the future.')

        self.get_state(commit=True)
        self.invoice_status = self.INVOICE_STATUS_PAID
        self.clean()

        if commit:
            self.save()
            self.migrate_state(
                user_model=user_model,
                entity_slug=entity_slug,
                force_migrate=True,
                je_timestamp=date_paid
            )
            self.lock_ledger(commit=True)
        invoice_status_paid.send_robust(sender=self.__class__,
                                        instance=self,
                                        commited=commit,
                                        **kwargs)

    def get_mark_as_paid_html_id(self):
        """
        InvoiceModel Mark as Paid HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String
        """
        return f'djl-{self.uuid}-invoice-mark-as-paid'

    def get_mark_as_paid_url(self, entity_slug: str = None):
        """
        InvoiceModel Mark-as-Paid action URL.

        Parameters
        __________
        entity_slug: str
            Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
            is not cached on QuerySet.

        Returns
        _______
        str
            InvoiceModel mark-as-paid action URL.
        """
        if not entity_slug:
            entity_slug = self.ledger.entity.slug
        return reverse('django_ledger:invoice-action-mark-as-paid',
                       kwargs={
                           'entity_slug': entity_slug,
                           'invoice_pk': self.uuid
                       })

    def get_mark_as_paid_message(self):
        """
        Internationalized confirmation message with Invoice Number.

        Returns
        _______
        str
            Mark-as-Paid InvoiceModel confirmation message as a String.
        """
        return _('Do you want to mark Invoice %s as Paid?') % self.invoice_number

    # VOID...
    def mark_as_void(self,
                     entity_slug: str,
                     user_model,
                     date_void: Optional[Union[date, datetime]] = None,
                     commit: bool = False,
                     **kwargs):
        """
        Marks InvoiceModel as Void.
        When mark as void, all transactions associated with InvoiceModel are reversed as of the void date.

        Parameters
        __________

        entity_slug: str
            Entity slug associated with the InvoiceModel. Avoids additional DB query if passed.

        user_model
            UserModel associated with request.

        date_void: date
            InvoiceModel void date. Defaults to localdate() if None.

        commit: bool
            Commits transaction into DB. Defaults to False.
        """
        if not self.can_void():
            raise InvoiceModelValidationError(f'Cannot mark Invoice {self.uuid} as Void...')

        if date_void:
            if isinstance(date_void, datetime):
                self.date_void = date_void.date()
            elif isinstance(date_void, date):
                self.date_void = date_void
        else:
            self.date_void = get_localdate()

        if self.date_void > get_localdate():
            raise InvoiceModelValidationError(f'Cannot void {self.__class__.__name__} in the future.')
        if self.date_void < self.date_approved:
            raise InvoiceModelValidationError(
                f'Cannot void {self.__class__.__name__} at {self.date_void} before approved '
                f'{self.date_approved}')

        self.void_state(commit=True)
        self.invoice_status = self.INVOICE_STATUS_VOID
        self.clean()

        if commit:
            self.unlock_ledger(commit=True, raise_exception=False)
            self.migrate_state(
                user_model=user_model,
                entity_slug=entity_slug,
                void=True,
                void_date=self.date_void,
                force_migrate=True,
                raise_exception=False
            )
            self.save()
            self.lock_ledger(commit=True, raise_exception=False)
        invoice_status_void.send_robust(sender=self.__class__,
                                        instance=self,
                                        commited=commit,
                                        **kwargs)

    def get_mark_as_void_html_id(self):
        """
        InvoiceModel Mark as Void HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-{self.uuid}-invoice-mark-as-void'

    def get_mark_as_void_url(self):
        """
        InvoiceModel Mark-as-Void action URL.

        Parameters
        __________
        entity_slug: str
            Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
            is not cached on QuerySet.

        Returns
        _______
            InvoiceModel mark-as-void action URL.
        """
        return reverse('django_ledger:invoice-action-mark-as-void',
                       kwargs={
                           'entity_slug': self.ledger.entity.slug,
                           'invoice_pk': self.uuid
                       })

    def get_mark_as_void_message(self):
        """
        Internationalized confirmation message with Invoice Number.

        Returns
        _______
        str
            Mark-as-Void InvoiceModel confirmation message as a String.
        """
        return _('Do you want to mark Invoice %s as Void?') % self.invoice_number

    # CANCEL
    def mark_as_canceled(self,
                         date_canceled: date = None,
                         commit: bool = False,
                         **kwargs):
        """
        Mark InvoiceModel as Canceled.

        Parameters
        __________

        date_canceled: date
            InvoiceModel canceled date. Defaults to localdate() if None.

        commit: bool
            Commits transaction into the Database. Defaults to False.
        """
        if not self.can_cancel():
            raise InvoiceModelValidationError(f'Cannot cancel Invoice {self.invoice_number}.')

        self.date_canceled = get_localdate() if not date_canceled else date_canceled
        self.invoice_status = self.INVOICE_STATUS_CANCELED
        self.clean()
        if commit:
            self.unlock_ledger(commit=True, raise_exception=False)
            self.unpost_ledger(commit=True, raise_exception=False)
            self.save()
        invoice_status_canceled.send_robust(sender=self.__class__,
                                            instance=self,
                                            commited=commit,
                                            **kwargs)

    def get_mark_as_canceled_html_id(self):
        """
        InvoiceModel Mark as Canceled HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-{self.uuid}-invoice-mark-as-canceled'

    def get_mark_as_canceled_url(self):
        """
        InvoiceModel Mark-as-Canceled action URL.

        Parameters
        __________
        entity_slug: str
            Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
            is not cached on QuerySet.

        Returns
        _______
        str
            InvoiceModel mark-as-canceled action URL.
        """
        return reverse('django_ledger:invoice-action-mark-as-canceled',
                       kwargs={
                           'entity_slug': self.ledger.entity.slug,
                           'invoice_pk': self.uuid
                       })

    def get_mark_as_canceled_message(self):
        """
        Internationalized confirmation message with Invoice Number.

        Returns
        _______
        str
            Mark-as-Canceled InvoiceModel confirmation message as a String.
        """
        return _('Do you want to mark Invoice %s as Canceled?') % self.invoice_number

    # DELETE ACTIONS...
    def delete(self, force_db_delete: bool = False, using=None, keep_parents=False):
        if not force_db_delete:
            self.mark_as_canceled(commit=True)
            return
        if not self.can_delete():
            raise InvoiceModelValidationError(
                message=_(f'Invoice {self.invoice_number} cannot be deleted...')
            )
        return super().delete(using=using, keep_parents=keep_parents)

    # def mark_as_delete(self, **kwargs):
    #     """
    #     Deletes InvoiceModel from DB if possible. Raises exception if can_delete() is False.
    #     """
    #     if not self.can_delete():
    #         raise InvoiceModelValidationError(
    #             f'Invoice {self.invoice_number} cannot be deleted. Must be void after Approved.')
    #     self.delete(**kwargs)
    #
    # # def get_mark_as_delete_html_id(self) -> str:
    #     """
    #     InvoiceModel Mark as Delete URL.
    #
    #     Returns
    #     _______
    #
    #     str
    #         URL as a String.
    #     """
    #     return f'djl-invoice-model-{self.uuid}-mark-as-delete'
    #
    # def get_mark_as_delete_url(self, entity_slug: Optional[str] = None) -> str:
    #     """
    #     InvoiceModel Mark-as-Delete action URL.
    #
    #     Parameters
    #     __________
    #     entity_slug: str
    #         Entity Slug kwarg. If not provided, will result in addition DB query if select_related('ledger__entity')
    #         is not cached on QuerySet.
    #
    #     Returns
    #     _______
    #     str
    #         InvoiceModel mark-as-delete action URL.
    #     """
    #     if not entity_slug:
    #         entity_slug = self.ledger.entity.slug
    #     return reverse('django_ledger:invoice-action-mark-as-delete',
    #                    kwargs={
    #                        'entity_slug': entity_slug,
    #                        'invoice_pk': self.uuid
    #                    })
    #
    # def get_mark_as_delete_message(self) -> str:
    #     """
    #     Internationalized confirmation message with Invoice Number.
    #
    #     Returns
    #     _______
    #     str
    #         Mark-as-Delete InvoiceModel confirmation message as a String.
    #     """
    #     return _('Do you want to delete Invoice %s?') % self.invoice_number

    # ACTIONS END....
    def get_status_action_date(self):
        """
        Current status action date.

        Returns
        _______
        date
            A date. i.e. If status is Approved, return date_approved. If Paid, return date_paid.
        """
        return getattr(self, f'date_{self.invoice_status}')

    def get_document_id(self):
        """
        Human-readable document number. Defaults to invoice_number.

        Returns
        _______
        str
            Document Number as a String.
        """
        return self.invoice_number

    def get_html_id(self):
        """
        Unique InvoiceNumber HTML ID.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-invoice-model-{self.REL_NAME_PREFIX}-{self.uuid}'

    def get_html_amount_due_id(self):
        """
        Unique amount due HTML ID.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-invoice-model-{self.REL_NAME_PREFIX}-{self.uuid}-amount-due'

    def get_html_amount_paid_id(self):
        """
        Unique amount paid HTML ID

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-invoice-model-{self.REL_NAME_PREFIX}-{self.uuid}-amount-paid'

    def get_html_form_id(self):
        """
        Unique InvoiceModel Form HTML ID.

        Returns
        _______
        str
            HTML ID as a String.
        """
        return f'djl-invoice-model-{self.REL_NAME_PREFIX}-{self.uuid}-form'

    def get_terms_start_date(self) -> date:
        """
        Date where InvoiceModel term start to apply.

        Returns
        _______
        date
            A date which represents the start of InvoiceModel terms.
        """
        return self.date_approved

    def _get_next_state_model(self, raise_exception=True):
        """
        Fetches the next sequenced state model associated with the InvoiceModel number.

        Parameters
        ----------
        raise_exception: bool
            Raises IntegrityError if unable to secure transaction from DB.

        Returns
        -------
        EntityStateModel
            An instance of EntityStateModel
        """
        EntityStateModel = lazy_loader.get_entity_state_model()
        EntityModel = lazy_loader.get_entity_model()
        entity_model = EntityModel.objects.get(uuid__exact=self.ledger.entity_id)
        fy_key = entity_model.get_fy_for_date(dt=self.date_draft)
        try:
            LOOKUP = {
                'entity_model_id__exact': self.ledger.entity_id,
                'entity_unit_id__exact': None,
                'fiscal_year': fy_key,
                'key__exact': EntityStateModel.KEY_INVOICE
            }

            state_model_qs = EntityStateModel.objects.filter(**LOOKUP).select_related(
                'entity_model').select_for_update()
            state_model = state_model_qs.get()
            state_model.sequence = F('sequence') + 1
            state_model.save()
            state_model.refresh_from_db()
            return state_model
        except ObjectDoesNotExist:
            EntityModel = lazy_loader.get_entity_model()
            entity_model = EntityModel.objects.get(uuid__exact=self.ledger.entity_id)
            fy_key = entity_model.get_fy_for_date(dt=self.date_draft)

            LOOKUP = {
                'entity_model_id': entity_model.uuid,
                'entity_unit_id': None,
                'fiscal_year': fy_key,
                'key': EntityStateModel.KEY_INVOICE,
                'sequence': 1
            }

            state_model = EntityStateModel.objects.create(**LOOKUP)
            return state_model
        except IntegrityError as e:
            if raise_exception:
                raise e

    def generate_invoice_number(self, commit: bool = False) -> str:
        """
        Atomic Transaction. Generates the next InvoiceModel document number available. The operation
        will result in two additional queries if the InvoiceModel & LedgerModel is not cached in
        QuerySet via select_related('ledger').

        Parameters
        __________
        commit: bool
            Commits transaction into InvoiceModel.

        Returns
        _______
        str
            A String, representing the generated InvoiceModel instance Document Number.
        """
        if self.can_generate_invoice_number():
            with transaction.atomic(durable=True):
                state_model = None
                while not state_model:
                    state_model = self._get_next_state_model(raise_exception=False)
                seq = str(state_model.sequence).zfill(DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING)
                self.invoice_number = f'{DJANGO_LEDGER_INVOICE_NUMBER_PREFIX}-{state_model.fiscal_year}-{seq}'

                if commit:
                    self.save(update_fields=['invoice_number'])
        return self.invoice_number

    def generate_descriptive_title(self) -> str:
        return f'Bill {self.invoice_number} | {self.get_invoice_status_display()} {self.get_status_action_date()} | {self.customer.customer_name}'

    # --> URLs <---
    def get_absolute_url(self):
        return reverse('django_ledger:invoice-detail',
                       kwargs={
                           'entity_slug': self.ledger.entity.slug,
                           'invoice_pk': self.uuid
                       })

    def clean(self, commit: bool = True):
        """
        Clean method for InvoiceModel. Results in a DB query if invoice number has not been generated and the
        InvoiceModel is eligible to generate an invoice_number.

        Parameters
        __________

        commit: bool
            If True, commits into DB the generated InvoiceModel number if generated.
        """

        super().clean()

        if self.cash_account.role != ASSET_CA_CASH:
            raise ValidationError(f'Cash account must be of role {ASSET_CA_CASH}.')
        if self.prepaid_account.role != ASSET_CA_RECEIVABLES:
            raise ValidationError(f'Prepaid account must be of role {ASSET_CA_RECEIVABLES}.')
        if self.unearned_account.role != LIABILITY_CL_DEFERRED_REVENUE:
            raise ValidationError(f'Unearned account must be of role {LIABILITY_CL_DEFERRED_REVENUE}.')

    def save(self, **kwargs):
        """
        Save method for InvoiceModel. Results in a DB query if invoice number has not been generated and the
        InvoiceModel is eligible to generate a invoice_number.
        """
        if self.can_generate_invoice_number():
            self.generate_invoice_number(commit=True)
        super(InvoiceModelAbstract, self).save(**kwargs)


class InvoiceModel(InvoiceModelAbstract):
    """
    Base Invoice Model from Abstract.
    """

    class Meta(InvoiceModelAbstract.Meta):
        abstract = False


def invoicemodel_presave(instance: InvoiceModel, **kwargs):
    if instance.can_generate_invoice_number():
        instance.generate_invoice_number(commit=False)

    if not instance.entity_model_id:
        instance.entity_model = instance.ledger.entity


pre_save.connect(receiver=invoicemodel_presave, sender=InvoiceModel)



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

The TransactionModel serves as the foundational accounting entity where all financial transactions are recorded.
Every transaction must be associated with a JournalEntryModel, which represents a collection
of related transactions. This strict association ensures that standalone TransactionModels—or orphan transactions—do not
exist, a constraint enforced at the database level.

Each transaction performs either a CREDIT or a DEBIT operation on the designated AccountModel, upholding standard
accounting principles. The TransactionModel API integrates the IOMixIn, a critical component for generating financial
statements. This mixin facilitates efficient querying and aggregation directly at the database level, eliminating the need
to load all TransactionModels into memory. This database-driven approach significantly improves performance and simplifies
the process of generating accurate financial reports.

The TransactionModel, together with the IOMixIn, is essential for ensuring seamless, efficient, and reliable
financial statement production in the Django Ledger framework.
"""
import warnings
from datetime import datetime, date
from typing import List, Union, Set
from uuid import uuid4, UUID

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator
from django.db import models
from django.db.models import Q, QuerySet, Manager, F
from django.db.models.signals import pre_save
from django.utils.translation import gettext_lazy as _

from django_ledger.io.io_core import validate_io_timestamp
from django_ledger.models import (
    AccountModel,
    BillModel,
    EntityModel,
    InvoiceModel,
    LedgerModel
)
from django_ledger.models.deprecations import deprecated_entity_slug_behavior
from django_ledger.models.mixins import CreateUpdateMixIn
from django_ledger.models.unit import EntityUnitModel
from django_ledger.models.utils import lazy_loader
from django_ledger.settings import DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR

UserModel = get_user_model()


class TransactionModelValidationError(ValidationError):
    pass


class TransactionModelQuerySet(QuerySet):
    """
    A custom QuerySet class tailored for `TransactionModel` objects. It includes a collection
    of methods to efficiently and safely retrieve and filter transactions from the database
    based on common use cases.
    """

    def for_user(self, user_model) -> 'TransactionModelQuerySet':
        """
        Filters transactions accessible to a specific user based on their permissions.

        Parameters
        ----------
        user_model : UserModel
            The user object for which the transactions should be filtered.

        Returns
        -------
        TransactionModelQuerySet
            A queryset containing transactions filtered by the user's access level.

        Description
        -----------
        - Returns all `TransactionModel` objects for superusers.
        - For regular users, it filters transactions where:
          - The user is an admin of the entity associated with the ledger in the transaction.
          - The user is a manager of the entity associated with the ledger in the transaction.
        """

        if user_model.is_superuser:
            return self

        return self.filter(
            Q(journal_entry__ledger__entity__admin=user_model) |
            Q(journal_entry__ledger__entity__managers__in=[user_model])
        )

    def posted(self) -> 'TransactionModelQuerySet':
        """
        Retrieves transactions that are part of a posted journal entry and ledger.

        A transaction is considered "posted" if:
        - It belongs to a journal entry marked as *posted*.
        - Its associated journal entry is part of a ledger marked as *posted*.

        Returns
        -------
        TransactionModelQuerySet
            A QuerySet containing only transactions that meet the "posted" criteria.
        """
        return self.filter(
            Q(journal_entry__posted=True) &
            Q(journal_entry__ledger__posted=True)
        )

    def for_accounts(self, account_list: List[Union[AccountModel, str, UUID]]) -> 'TransactionModelQuerySet':
        """
        Filters transactions based on the accounts they are associated with.

        Parameters
        ----------
        account_list : list of str or AccountModel
            A list containing account codes (strings) or `AccountModel` instances.
            Transactions will be filtered to match these accounts.

        Returns
        -------
        TransactionModelQuerySet
            A QuerySet filtered for transactions associated with the specified accounts.
        """

        if not isinstance(account_list, list) or not len(account_list) > 0:
            raise TransactionModelValidationError(
                message=_('Account list must be a list of AccountModel, UUID or str objects (codes).')
            )
        if isinstance(account_list[0], str):
            return self.filter(account__code__in=account_list)
        elif isinstance(account_list[0], UUID):
            return self.filter(account__uuid__in=account_list)
        elif isinstance(account_list[0], AccountModel):
            return self.filter(account__in=account_list)
        raise TransactionModelValidationError(
            message=_('Account list must be a list of AccountModel, UUID or str objects (codes).')
        )

    def for_roles(self, role_list: Union[str, List[str], Set[str]]) -> 'TransactionModelQuerySet':
        """
        Fetches a QuerySet of TransactionModels which AccountModel has a specific role.

        Parameters
        ----------
        role_list: str or list
            A string or list of strings representing the roles to be used as filter.

        Returns
        -------
        TransactionModelQuerySet
            Returns a TransactionModelQuerySet with applied filters.
        """
        if isinstance(role_list, str):
            return self.filter(account__role__in=[role_list])
        return self.filter(account__role__in=role_list)

    def for_unit(self, unit_slug: Union[str, EntityUnitModel]) -> 'TransactionModelQuerySet':
        """
        Filters transactions based on their associated entity unit.

        Parameters
        ----------
        unit_slug : str or EntityUnitModel
            A string representing the slug of the entity unit or an `EntityUnitModel` instance.

        Returns
        -------
        TransactionModelQuerySet
            A QuerySet filtered for transactions linked to the specified unit.
        """
        if isinstance(unit_slug, EntityUnitModel):
            return self.filter(journal_entry__entity_unit=unit_slug)
        return self.filter(journal_entry__entity_unit__slug__exact=unit_slug)

    def for_activity(self, activity_list: Union[str, List[str], Set[str]]) -> 'TransactionModelQuerySet':
        """
        Filters transactions based on their associated activity or activities.

        Parameters
        ----------
        activity_list : str or list of str
            A single activity or a list of activities to filter transactions by.

        Returns
        -------
        TransactionModelQuerySet
            A QuerySet filtered for transactions linked to the specified activity or activities.
        """
        if isinstance(activity_list, str):
            return self.filter(journal_entry__activity__in=[activity_list])
        return self.filter(journal_entry__activity__in=activity_list)

    def to_date(self, to_date: Union[str, date, datetime]) -> 'TransactionModelQuerySet':
        """
        Filters transactions occurring on or before a specific date or timestamp.

        If `to_date` is a naive datetime (no timezone), it is assumed to be in local time
        based on Django settings.

        Parameters
        ----------
        to_date : str, date, or datetime
            The maximum date or timestamp for filtering. When using a date (not datetime),
            the filter is inclusive (e.g., "2022-12-20" includes all transactions from that day).

        Returns
        -------
        TransactionModelQuerySet
            A QuerySet filtered to include transactions up to the specified date or timestamp.
        """

        if isinstance(to_date, str):
            to_date = validate_io_timestamp(to_date)

        if isinstance(to_date, date):
            return self.filter(journal_entry__timestamp__date__lte=to_date)
        return self.filter(journal_entry__timestamp__lte=to_date)

    def from_date(self, from_date: Union[str, date, datetime]) -> 'TransactionModelQuerySet':
        """
        Filters transactions occurring on or after a specific date or timestamp.

        If `from_date` is a naive datetime (no timezone), it is assumed to be in local time
        based on Django settings.

        Parameters
        ----------
        from_date : str, date, or datetime
            The minimum date or timestamp for filtering. When using a date (not datetime),
            the filter is inclusive (e.g., "2022-12-20" includes all transactions from that day).

        Returns
        -------
        TransactionModelQuerySet
            A QuerySet filtered to include transactions from the specified date or timestamp onwards.
        """
        if isinstance(from_date, str):
            from_date = validate_io_timestamp(from_date)

        if isinstance(from_date, date):
            return self.filter(journal_entry__timestamp__date__gte=from_date)

        return self.filter(journal_entry__timestamp__gte=from_date)

    def not_closing_entry(self) -> 'TransactionModelQuerySet':
        """
        Filters transactions that are *not* part of a closing journal entry.

        Returns
        -------
        TransactionModelQuerySet
            A QuerySet with transactions where the `journal_entry__is_closing_entry` field is False.
        """
        return self.filter(journal_entry__is_closing_entry=False)

    def is_closing_entry(self) -> 'TransactionModelQuerySet':
        """
        Filters transactions that are part of a closing journal entry.

        Returns
        -------
        TransactionModelQuerySet
            A QuerySet with transactions where the `journal_entry__is_closing_entry` field is True.
        """
        return self.filter(journal_entry__is_closing_entry=True)

    def for_ledger(self, ledger_model: Union[LedgerModel, UUID, str]) -> 'TransactionModelQuerySet':
        """
        Filters transactions for a specific ledger under a given entity.

        Parameters
        ----------
        ledger_model : Union[LedgerModel, UUID]
            The ledger model or its UUID to filter by.

        Returns
        -------
        TransactionModelQuerySet
            A queryset containing transactions associated with the given ledger and entity.
        """
        if isinstance(ledger_model, UUID):
            return self.filter(journal_entry__ledger__uuid__exact=ledger_model)
        return self.filter(journal_entry__ledger=ledger_model)

    def for_journal_entry(self, je_model) -> 'TransactionModelQuerySet':
        """
        Filters transactions for a specific journal entry under a given ledger and entity.

        Parameters
        ----------
        je_model : Union[JournalEntryModel, UUID]
            The journal entry model or its UUID to filter by.

        Returns
        -------
        TransactionModelQuerySet
            A queryset containing transactions associated with the given journal entry.
        """
        if isinstance(je_model, lazy_loader.get_journal_entry_model()):
            return self.filter(journal_entry=je_model)
        return self.filter(journal_entry__uuid__exact=je_model)

    def for_bill(self, bill_model: Union[BillModel, str, UUID]) -> 'TransactionModelQuerySet':
        """
        Filters transactions for a specific bill under a given entity.

        Parameters
        ----------
        bill_model : Union[BillModel, str, UUID]
            The bill model or its UUID to filter by.

        Returns
        -------
        TransactionModelQuerySet
            A queryset containing transactions related to the specified bill.
        """
        if isinstance(bill_model, BillModel):
            return self.filter(journal_entry__ledger__billmodel=bill_model)
        return self.filter(journal_entry__ledger__billmodel__uuid__exact=bill_model)

    def for_invoice(self, invoice_model: Union[InvoiceModel, str, UUID]) -> 'TransactionModelQuerySet':
        """
        Filters transactions for a specific invoice under a given entity.

        Parameters
        ----------
        invoice_model : Union[InvoiceModel, str, UUID]
            The invoice model or its UUID to filter by.

        Returns
        -------
        TransactionModelQuerySet
            A queryset containing transactions related to the specified invoice.
        """
        if isinstance(invoice_model, InvoiceModel):
            return self.filter(journal_entry__ledger__invoicemodel=invoice_model)
        return self.filter(journal_entry__ledger__invoicemodel__uuid__exact=invoice_model)

    def with_annotated_details(self) -> 'TransactionModelQuerySet':
        return self.annotate(
            entity_unit_name=F('journal_entry__entity_unit__name'),
            account_code=F('account__code'),
            account_name=F('account__name'),
            timestamp=F('journal_entry__timestamp'),
        )

    def is_cleared(self) -> 'TransactionModelQuerySet':
        return self.filter(cleared=True)

    def not_cleared(self) -> 'TransactionModelQuerySet':
        return self.filter(cleared=False)

    def is_reconciled(self) -> 'TransactionModelQuerySet':
        return self.filter(reconciled=True)

    def not_reconciled(self) -> 'TransactionModelQuerySet':
        return self.filter(reconciled=False)


class TransactionModelManager(Manager):
    """
    A custom manager for `TransactionModel` designed to add helper methods for
    querying and filtering `TransactionModel` objects efficiently based on use cases like
    user permissions, associated entities, ledgers, journal entries, and more.

    This manager leverages `TransactionModelQuerySet` for complex query construction and
    integrates advanced filtering options based on user roles, entities, and other relationships.
    """

    def get_queryset(self) -> TransactionModelQuerySet:
        """
        Retrieves the base queryset for `TransactionModel`, annotated and pre-loaded
        with commonly used related fields.

        Returns
        -------
        TransactionModelQuerySet
            A custom queryset with essential annotations and relationships preloaded.
        """
        qs = TransactionModelQuerySet(self.model, using=self._db)
        return qs.annotate(
            timestamp=F('journal_entry__timestamp'),
            _coa_id=F('account__coa_model_id')  # Annotates the `coa_model_id` from the related `account`.
        ).select_related(
            'journal_entry',  # Pre-loads the related Journal Entry.
            'account',  # Pre-loads the Account associated with the Transaction.
            'account__coa_model',  # Pre-loads the Chart of Accounts related to the Account.
        )

    @deprecated_entity_slug_behavior
    def for_entity(self, entity_model: EntityModel | str | UUID = None, **kwargs) -> TransactionModelQuerySet:
        """
        Filters transactions for a specific entity, optionally scoped to a specific user.

        Parameters
        ----------
        entity_model : Union[EntityModel, str, UUID]
            Identifier for the entity. This can be an `EntityModel` object, a slug (str), or a UUID.

        Returns
        -------
        TransactionModelQuerySet
            A queryset containing transactions associated with the specified entity.

        Notes
        -----
        - If `user_model` is provided, only transactions accessible by the user are included.
        - Supports flexible filtering by accepting different forms of `entity_slug`.
        """

        qs = self.get_queryset()
        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(journal_entry__ledger__entity=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(journal_entry__ledger__entity_id=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(journal_entry__ledger__entity__slug__exact=entity_model)
        else:
            raise TransactionModelValidationError(
                message='entity_model parameter must be either an EntityModel, String or a UUID.'
            )
        return qs


class TransactionModelAbstract(CreateUpdateMixIn):
    """
    Abstract base model class for representing financial transactions in a ledger system.

    This class defines the structure and behavior of transactions in a generic ledger system,
    including attributes such as transaction type, associated account, amount, description,
    and status flags for cleared and reconciled state. It also includes methods and properties
    to facilitate queries and determine transaction type.

    Attributes
    ----------
    uuid : UUID
        Unique identifier of the transaction.
    tx_type : str
        Type of the transaction, either 'credit' or 'debit'.
    journal_entry : ForeignKey
        Reference to the associated JournalEntryModel instance for this transaction.
    account : ForeignKey
        Reference to the associated account in the Chart of Accounts.
    amount : Decimal
        Monetary amount for the transaction. Must be a non-negative value.
    description : str, optional
        Description of the transaction.
    cleared : bool
        Indicates if the transaction has been cleared. Defaults to False.
    reconciled : bool
        Indicates if the transaction has been reconciled. Defaults to False.
    objects : TransactionModelManager
        Custom manager for managing transaction models.
    """

    CREDIT = 'credit'
    DEBIT = 'debit'
    TX_TYPE = [
        (CREDIT, _('Credit')),
        (DEBIT, _('Debit'))
    ]

    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    tx_type = models.CharField(max_length=10, choices=TX_TYPE, verbose_name=_('Transaction Type'))

    journal_entry = models.ForeignKey(
        'django_ledger.JournalEntryModel',
        editable=False,
        verbose_name=_('Journal Entry'),
        help_text=_('Journal Entry to be associated with this transaction.'),
        on_delete=models.CASCADE
    )
    account = models.ForeignKey(
        'django_ledger.AccountModel',
        verbose_name=_('Account'),
        help_text=_('Account from Chart of Accounts to be associated with this transaction.'),
        on_delete=models.PROTECT
    )
    amount = models.DecimalField(
        decimal_places=2,
        max_digits=20,
        default=0.00,
        verbose_name=_('Amount'),
        help_text=_('Amount of the transaction.'),
        validators=[MinValueValidator(0)]
    )
    description = models.CharField(
        max_length=100,
        null=True,
        blank=True,
        verbose_name=_('Transaction Description'),
        help_text=_('A description to be included with this individual transaction.')
    )
    cleared = models.BooleanField(default=False, verbose_name=_('Cleared'))
    reconciled = models.BooleanField(default=False, verbose_name=_('Reconciled'))
    objects = TransactionModelManager.from_queryset(TransactionModelQuerySet)()

    class Meta:
        abstract = True
        ordering = ['-created']
        verbose_name = _('Transaction')
        verbose_name_plural = _('Transactions')
        indexes = [
            models.Index(fields=['tx_type']),
            models.Index(fields=['account']),
            models.Index(fields=['journal_entry']),
            models.Index(fields=['created']),
            models.Index(fields=['updated']),
            models.Index(fields=['cleared']),
            models.Index(fields=['reconciled']),
        ]

    def __str__(self):
        return '{code}-{name}/{balance_type}: {amount}/{tx_type}'.format(
            code=self.account.code,
            name=self.account.name,
            balance_type=self.account.balance_type,
            amount=self.amount,
            tx_type=self.tx_type
        )

    @property
    def coa_id(self):
        """
        Returns the Chart of Accounts (COA) ID associated with the current object.

        The property attempts to retrieve the COA ID value stored internally, if it exists.
        If the internal value is not set and the `account` attribute is not `None`,
        the COA ID is obtained from the `account.coa_model_id` attribute. Otherwise,
        it will return `None`.

        Returns
        -------
        Any or None
            The COA ID if it exists, otherwise `None`.
        """
        try:
            return getattr(self, '_coa_id')
        except AttributeError:
            if self.account is None:
                return None
            return self.account.coa_model_id

    def is_debit(self):
        """
        Determines if the transaction type is a debit.

        This method checks whether the transaction type of the current instance is
        set to a debit type.

        Returns
        -------
        bool
            True if the transaction type is debit, otherwise False.
        """
        return self.tx_type == self.DEBIT

    def is_credit(self):
        """
        Check if the transaction is of type 'credit'.

        This method evaluates whether the transaction type of the current instance
        matches the 'credit' transaction type.

        Returns
        -------
        bool
            True if the transaction type is 'credit', False otherwise.
        """
        return self.tx_type == self.CREDIT


class TransactionModel(TransactionModelAbstract):
    """
    Base Transaction Model From Abstract.
    """

    class Meta(TransactionModelAbstract.Meta):
        abstract = False


def transactionmodel_presave(instance: TransactionModel, **kwargs):
    """
    Pre-save validation for the TransactionModel instance.

    This function is executed before saving a `TransactionModel` instance,
    ensuring that certain conditions are met to maintain data integrity.

    The function performs the following validations:
    1. **Account Transactionality**:
       If the `bypass_account_state` flag is not provided or set to `False`,
       it verifies whether the associated account can process transactions
       by calling `instance.account.can_transact()`. If the account cannot
       process transactions, the save operation is interrupted to prevent
       invalid data.

    2. **Journal Entry Lock**:
       If the associated journal entry (`instance.journal_entry`) is locked,
       the transaction cannot be modified. The save process is halted if the
       journal entry is marked as locked.

    Parameters
    ----------
    instance : TransactionModel
        The `TransactionModel` instance that is about to be saved.
    kwargs : dict
        Additional keyword arguments, such as the optional `bypass_account_state`.

    Raises
    ------
    TransactionModelValidationError
        Raised in the following scenarios:
        - **Account Transactionality Failure**:
          When `bypass_account_state` is `False` or not provided, and the
          associated account (`instance.account`) cannot process transactions.
          The exception contains a message identifying the account.

        - **Locked Journal Entry**:
          When the associated journal entry (`instance.journal_entry`) is locked,
          preventing modification of any related transactions. The error message
          describes the locked journal entry constraint.
    """
    bypass_account_state = kwargs.get('bypass_account_state', False)

    if instance.account_id and instance.account.is_root_account():
        raise TransactionModelValidationError(
            message=_('Transactions cannot be linked to root accounts.')
        )

    if all([
        not bypass_account_state,
        not instance.account.can_transact()
    ]):
        raise TransactionModelValidationError(
            message=_(f'Cannot create or modify transactions on account model {instance.account}.')
        )
    if instance.journal_entry.is_locked():
        raise TransactionModelValidationError(
            message=_('Cannot modify transactions on locked journal entries.')
        )


pre_save.connect(transactionmodel_presave, sender=TransactionModel)



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

An EntityUnit is a logical, user-defined grouping assigned to JournalEntryModels,
helping to segregate business operations into distinct components. Examples of
EntityUnits may include departments (e.g., Human Resources, IT), office locations,
real estate properties, or any other labels relevant to the business.

EntityUnits are self-contained entities, meaning that double-entry accounting rules
apply to all transactions associated with them. When an Invoice or Bill is updated,
the migration process generates the corresponding Journal Entries for each relevant
unit. This allows invoices or bills to split specific items into different units,
with the migration process allocating costs to each unit accordingly.

Key advantages of EntityUnits:
    1. EntityUnits can generate their own financial statements, providing deeper
       insights into the specific operations of the business.
    2. EntityUnits can be assigned to specific items on Bills and Invoices, offering
       flexibility to track inventory, expenses, or income associated with distinct
       business units.
"""
import warnings
from random import choices
from string import ascii_lowercase, digits, ascii_uppercase
from typing import Optional
from uuid import uuid4, UUID

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q, F
from django.urls import reverse
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _
from treebeard.mp_tree import MP_Node, MP_NodeManager, MP_NodeQuerySet

from django_ledger.io.io_core import IOMixIn
from django_ledger.models import lazy_loader
from django_ledger.models.deprecations import deprecated_entity_slug_behavior
from django_ledger.models.mixins import CreateUpdateMixIn, SlugNameMixIn
from django_ledger.settings import DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR

ENTITY_UNIT_RANDOM_SLUG_SUFFIX = ascii_lowercase + digits


class EntityUnitModelValidationError(ValidationError):
    pass


class EntityUnitModelQuerySet(MP_NodeQuerySet):

    def for_user(self, user_model) -> 'EntityUnitModelQuerySet':
        if user_model.is_superuser:
            return self
        return self.filter(
            Q(entity__admin=user_model) |
            Q(entity__managers__in=[user_model])
        )


class EntityUnitModelManager(MP_NodeManager):

    def get_queryset(self):
        qs = EntityUnitModelQuerySet(self.model, using=self._db)
        return qs.annotate(
            _entity_slug=F('entity__slug'),
            _entity_name=F('entity__name'),
        )

    @deprecated_entity_slug_behavior
    def for_entity(self, entity_model: 'EntityModel | str | UUID' = None, **kwargs):
        """
        Filter the queryset based on the provided entity model, its slug, or its UUID.

        Provides functionality to filter entities within a queryset by comparing either
        an EntityModel instance, its slug, or its UUID. This method also handles optional
        deprecated behavior for filtering by the 'user_model' for backward compatibility.

        Parameters
        ----------
        entity_model : EntityModel | str | UUID
            An entity filter criterion that could be a specific EntityModel instance, a
            string representing the entity slug, or a UUID corresponding to the entity.

        **kwargs : dict, optional
            Additional keyword arguments. If the 'user_model' parameter is provided,
            a deprecation warning is issued, and the functionality temporarily delegates
            to legacy behavior based on the value of 'DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR'.

        Returns
        -------
        QuerySet
            A queryset filtered based on the provided entity criteria.

        Raises
        ------
        EntityUnitModelValidationError
            Raised when the `entity_model` parameter does not match any of the accepted types
            (EntityModel, str, or UUID) and fails validation.
        """
        EntityModel = lazy_loader.get_entity_model()

        qs = self.get_queryset()
        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(entity=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(entity__slug__exact=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(entity_id=entity_model)
        else:
            raise EntityUnitModelValidationError(
                message='Must pass EntityModel, slug or UUID'
            )
        return qs


class EntityUnitModelAbstract(MP_Node,
                              IOMixIn,
                              SlugNameMixIn,
                              CreateUpdateMixIn):
    """
    Base implementation of the EntityUnitModel.

    Attributes
    ----------
    uuid : UUID
        This is a unique primary key generated for the table. The default value of this field is uuid4().

    slug: str
        A unique, indexed identifier for the EntityUnitModel instance used in URLs and queries.

    entity: EntityModel
        The EntityModel associated with this EntityUnitModel.

    document_prefix: str
        A predefined prefix automatically incorporated into JournalEntryModel document numbers. Max Length 3.
        May be user defined. Must be unique for the EntityModel.

    active: bool
        Active EntityUnits may transact. Inactive units are considered archived. Defaults to True.

    hidden: bool
        Hidden Units will not show on drop down menus on the UI. Defaults to False.
    """
    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    slug = models.SlugField(max_length=50)
    entity = models.ForeignKey('django_ledger.EntityModel',
                               editable=False,
                               on_delete=models.CASCADE,
                               verbose_name=_('Unit Entity'))
    document_prefix = models.CharField(max_length=3)
    active = models.BooleanField(default=True, verbose_name=_('Is Active'))
    hidden = models.BooleanField(default=False, verbose_name=_('Is Hidden'))

    objects = EntityUnitModelManager.from_queryset(queryset_class=EntityUnitModelQuerySet)()

    class Meta:
        abstract = True
        ordering = ['-created']
        verbose_name = _('Entity Unit Model')
        unique_together = [
            ('entity', 'slug'),
            ('entity', 'document_prefix'),
        ]
        indexes = [
            models.Index(fields=['active']),
            models.Index(fields=['hidden']),
            models.Index(fields=['entity']),
        ]

    def __str__(self):
        return f'{self.entity_name}: {self.name}'

    @property
    def entity_slug(self):
        try:
            return getattr(self, '_entity_slug')
        except AttributeError:
            pass
        return self.entity.slug

    @property
    def entity_name(self):
        try:
            return getattr(self, '_entity_name')
        except AttributeError:
            pass
        return self.entity.name

    def clean(self):
        self.create_entity_unit_slug()

        if not self.document_prefix:
            self.document_prefix = ''.join(choices(ascii_uppercase, k=3))

    def get_dashboard_url(self) -> str:
        """
        The dashboard URL of the EntityModelUnit.

        Returns
        -------
        str
            The EntityModelUnit instance dashboard URL.
        """
        return reverse('django_ledger:unit-dashboard',
                       kwargs={
                           'entity_slug': self.slug
                       })

    def get_entity_name(self) -> str:
        return self.entity.name

    def create_entity_unit_slug(self,
                                name: Optional[str] = None,
                                force: bool = False,
                                add_suffix: bool = True,
                                k: int = 5) -> str:
        """
        Automatically generates a EntityUnitModel slug. If slug is present, will not be replaced.
        Called during the clean() method.

        Parameters
        ----------
        force: bool
            Forces generation of new slug if already present.
        name: str
            The name used to create slug. If none, the unit name will be used.
        add_suffix: bool
            Adds a random suffix to the slug. Defaults to True.
        k: int
            Length of the suffix if add_suffix is True. Defaults to 5.

        Returns
        -------
        str
            The EntityUnitModel slug, regardless if generated or not.
        """
        if not self.slug or force:
            if not name:
                name = f'{self.name} Unit'
            unit_slug = slugify(name)
            if add_suffix:
                suffix = ''.join(choices(ENTITY_UNIT_RANDOM_SLUG_SUFFIX, k=k))
                unit_slug = f'{unit_slug}-{suffix}'
            self.slug = unit_slug
        return self.slug

    def get_absolute_url(self):
        return reverse(
            viewname='django_ledger:unit-detail',
            kwargs={
                'entity_slug': self.entity.slug,
                'unit_slug': self.slug
            }
        )


class EntityUnitModel(EntityUnitModelAbstract):
    """
    Base Model Class for EntityUnitModel
    """

    class Meta(EntityUnitModelAbstract.Meta):
        abstract = False



from django import forms

from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES
from django.utils.translation import gettext_lazy as _


class BugReportForm(forms.Form):
    DEVICE_CHOICES = [
        ('desktop', _('Desktop')),
        ('tablet', _('Tablet')),
        ('mobile', _('Mobile'))
    ]

    reproduce = forms.CharField(
        label=_('How to reproduce?'),
        required=True,
        widget=forms.Textarea(attrs={
            'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' textarea',
            'rows': 3
        }))

    expectation = forms.CharField(
        label=_('What did you expect?'),
        required=True,
        widget=forms.Textarea(attrs={
            'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' textarea',
            'rows': 3
        }))

    device = forms.ChoiceField(
        choices=DEVICE_CHOICES,
        widget=forms.Select(attrs={
            'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
        }),
        required=True,
        label=_('What device are you using?'))


class RequestNewFeatureForm(forms.Form):
    feature_description = forms.CharField(
        label=_('Is your feature request related to a problem? Please describe.'),
        required=True,
        widget=forms.Textarea(attrs={
            'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' textarea',
            'rows': 3
        }))

    solution = forms.CharField(
        label=_('Describe the solution you\'d like'),
        required=True,
        widget=forms.Textarea(attrs={
            'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' textarea',
            'rows': 3
        }))

    alternatives = forms.CharField(
        label=_('Describe alternatives you\'ve considered'),
        widget=forms.Textarea(attrs={
            'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' textarea',
            'rows': 3
        }))



from datetime import timedelta, datetime
from random import randint
from zoneinfo import ZoneInfo

from django.conf import settings

from django_ledger.io.io_core import IOValidationError
from django_ledger.models import EntityModel
from django_ledger.tests.base import DjangoLedgerBaseTest


class IOTest(DjangoLedgerBaseTest):

    def test_digest_dttm__dttm(self):
        self.assertTrue(settings.USE_TZ, msg='Timezone not enabled.')

        entity_model = self.get_random_entity_model()
        from_datetime = self.START_DATE
        to_datetime = self.START_DATE + timedelta(days=randint(10, 60))

        io_digest = entity_model.digest(from_date=from_datetime, to_date=to_datetime)
        self.assertTrue(isinstance(io_digest.get_to_datetime(), datetime))
        self.assertTrue(isinstance(io_digest.get_from_datetime(), datetime))
        self.assertEqual(io_digest.get_from_datetime(), from_datetime)
        self.assertEqual(io_digest.get_to_datetime(), to_datetime)

    def test_digest_dt__dttm(self):
        self.assertTrue(settings.USE_TZ, msg='Timezone not enabled.')

        entity_model = self.get_random_entity_model()
        from_datetime = self.START_DATE
        to_datetime = self.START_DATE + timedelta(days=randint(10, 60))

        io_digest = entity_model.digest(from_date=from_datetime.date(), to_date=to_datetime)
        self.assertTrue(isinstance(io_digest.get_to_datetime(), datetime))
        self.assertTrue(isinstance(io_digest.get_from_datetime(), datetime))

        self.assertEqual(
            # the assumed datetime given a date...
            io_digest.get_from_datetime(),

            # equals the localized datetime @ 0:00
            datetime.combine(
                from_datetime.date(),
                datetime.min.time(),
                tzinfo=ZoneInfo(settings.TIME_ZONE)
            )
        )

        self.assertEqual(io_digest.get_to_datetime(), to_datetime)

    def test_digest_dttm__dt(self):
        self.assertTrue(settings.USE_TZ, msg='Timezone not enabled.')

        entity_model = self.get_random_entity_model()
        from_datetime = self.START_DATE
        to_datetime = self.START_DATE + timedelta(days=randint(10, 60))

        io_digest = entity_model.digest(from_date=from_datetime, to_date=to_datetime.date())

        self.assertTrue(isinstance(io_digest.get_to_datetime(), datetime))
        self.assertTrue(isinstance(io_digest.get_from_datetime(), datetime))

        self.assertEqual(io_digest.get_from_datetime(), from_datetime)

        self.assertEqual(
            # the assumed datetime given a date...
            io_digest.get_to_datetime(),

            # equals the localized datetime @ 0:00
            datetime.combine(
                to_datetime.date(),
                datetime.min.time(),
                tzinfo=ZoneInfo(settings.TIME_ZONE)
            )
        )

    def test_digest_dt__dt(self):
        self.assertTrue(settings.USE_TZ, msg='Timezone not enabled.')

        entity_model = self.get_random_entity_model()
        from_datetime = self.START_DATE
        to_datetime = self.START_DATE + timedelta(days=randint(10, 60))

        io_digest = entity_model.digest(from_date=from_datetime.date(), to_date=to_datetime.date())

        self.assertTrue(isinstance(io_digest.get_to_datetime(), datetime))
        self.assertTrue(isinstance(io_digest.get_from_datetime(), datetime))

        self.assertEqual(
            # the assumed datetime given a date...
            io_digest.get_from_datetime(),

            # equals the localized datetime @ 0:00
            datetime.combine(
                from_datetime.date(),
                datetime.min.time(),
                tzinfo=ZoneInfo(settings.TIME_ZONE)
            )
        )

        self.assertEqual(
            # the assumed datetime given a date...
            io_digest.get_to_datetime(),

            # equals the localized datetime @ 0:00
            datetime.combine(
                to_datetime.date(),
                datetime.min.time(),
                tzinfo=ZoneInfo(settings.TIME_ZONE)
            )
        )

    def test_digest_none__dttm(self):
        self.assertTrue(settings.USE_TZ, msg='Timezone not enabled.')

        entity_model = self.get_random_entity_model()
        to_datetime = self.START_DATE + timedelta(days=randint(10, 60))

        io_digest = entity_model.digest(to_date=to_datetime)

        self.assertTrue(io_digest.get_from_datetime() is None)
        self.assertTrue(isinstance(io_digest.get_to_datetime(), datetime))

        self.assertEqual(
            # the assumed datetime given a date...
            io_digest.get_from_datetime(),

            # equals the localized datetime @ 0:00
            None
        )

        self.assertEqual(io_digest.get_to_datetime(), to_datetime)

    def test_digest_none__dt(self):
        self.assertTrue(settings.USE_TZ, msg='Timezone not enabled.')

        entity_model = self.get_random_entity_model()
        to_datetime = self.START_DATE + timedelta(days=randint(10, 60))

        io_digest = entity_model.digest(to_date=to_datetime.date())

        self.assertTrue(io_digest.get_from_datetime() is None)
        self.assertTrue(isinstance(io_digest.get_to_datetime(), datetime))

        self.assertEqual(
            # the assumed datetime given a date...
            io_digest.get_from_datetime(),

            # equals the localized datetime @ 0:00
            None
        )

        self.assertEqual(
            # the assumed datetime given a date...
            io_digest.get_to_datetime(),

            # equals the localized datetime @ 0:00
            datetime.combine(
                to_datetime.date(),
                datetime.min.time(),
                tzinfo=ZoneInfo(settings.TIME_ZONE)
            )
        )

    def test_digest_entity(self):
        entity_model = self.get_random_entity_model()
        from_datetime = self.START_DATE
        to_datetime = self.START_DATE + timedelta(days=randint(10, 60))

        with self.assertRaises(IOValidationError):
            entity_model.digest(
                entity_slug='1234',
                from_date=from_datetime,
                to_date=to_datetime
            )

        io_digest = entity_model.digest(
            entity_slug=entity_model.slug,
            from_date=from_datetime,
            to_date=to_datetime
        )

        self.assertTrue(isinstance(io_digest.IO_MODEL, EntityModel))
        self.assertTrue(io_digest.get_io_data(), io_digest.IO_DATA)
        self.assertTrue(io_digest.IO_DATA['entity_slug'], entity_model.slug)
        self.assertFalse(io_digest.IO_DATA['by_activity'])
        self.assertFalse(io_digest.IO_DATA['by_unit'])
        self.assertFalse(io_digest.IO_DATA['by_tx_type'])

        # io_digest = entity_model.digest(
        #     unit_slug='3212',
        #     from_date=from_datetime,
        #     to_date=to_datetime
        # )
        #
        # self.assertEqual(io_digest.get_io_txs_queryset().count(), 0)

    def test_io_transactions_belong_to_entity(self):
        entity_model = self.get_random_entity_model()
        from_datetime = self.START_DATE
        to_datetime = self.START_DATE + timedelta(days=randint(10, 60))

        io_digest = entity_model.digest(
            entity_slug=entity_model.slug,
            from_date=from_datetime,
            to_date=to_datetime,
            for_test=True
        )

        tx_qs = io_digest.get_io_txs_queryset()
        # Every transaction returned by the IO for an entity digest must belong to that entity.
        for tx in tx_qs:
            self.assertEqual(tx['journal_entry__ledger__entity_id'], entity_model.uuid)



from django.urls import path

from django_ledger import views

urlpatterns = [
    path('login/', views.DjangoLedgerLoginView.as_view(), name='login'),
    path('logout/', views.DjangoLedgerLogoutView.as_view(), name='logout'),
]



from django.urls import path

from django_ledger.views import (EstimateModelListView, EstimateModelCreateView,
                                 EstimateModelDetailView, EstimateModelUpdateView,
                                 EstimateActionMarkAsDraftView, EstimateActionMarkAsReviewView,
                                 EstimateActionMarkAsApprovedView, EstimateActionMarkAsCompletedView,
                                 EstimateActionMarkAsCanceledView)

urlpatterns = [
    path('<slug:entity_slug>/list/', EstimateModelListView.as_view(), name='customer-estimate-list'),
    path('<slug:entity_slug>/create/', EstimateModelCreateView.as_view(), name='customer-estimate-create'),
    path('<slug:entity_slug>/detail/<uuid:ce_pk>/',
         EstimateModelDetailView.as_view(),
         name='customer-estimate-detail'),
    path('<slug:entity_slug>/update/<uuid:ce_pk>/',
         EstimateModelUpdateView.as_view(),
         name='customer-estimate-update'),
    path('<slug:entity_slug>/update/<uuid:ce_pk>/items/',
         EstimateModelUpdateView.as_view(action_update_items=True),
         name='customer-estimate-update-items'),

    # Actions....
    path('<slug:entity_slug>/action/<uuid:ce_pk>/mark-as-draft/',
         EstimateActionMarkAsDraftView.as_view(),
         name='customer-estimate-action-mark-as-draft'),
    path('<slug:entity_slug>/action/<uuid:ce_pk>/mark-as-review/',
         EstimateActionMarkAsReviewView.as_view(),
         name='customer-estimate-action-mark-as-review'),
    path('<slug:entity_slug>/action/<uuid:ce_pk>/mark-as-approved/',
         EstimateActionMarkAsApprovedView.as_view(),
         name='customer-estimate-action-mark-as-approved'),
    path('<slug:entity_slug>/action/<uuid:ce_pk>/mark-as-completed/',
         EstimateActionMarkAsCompletedView.as_view(),
         name='customer-estimate-action-mark-as-completed'),
    path('<slug:entity_slug>/action/<uuid:ce_pk>/mark-as-canceled/',
         EstimateActionMarkAsCanceledView.as_view(),
         name='customer-estimate-action-mark-as-canceled'),
]



from typing import Union
from uuid import UUID

from django.forms import ModelForm, Textarea, Select, DateTimeInput, ValidationError
from django.utils.translation import gettext_lazy as _

from django_ledger.models import EntityModel
from django_ledger.models.journal_entry import JournalEntryModel
from django_ledger.models.ledger import LedgerModel
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class JournalEntryModelCreateForm(ModelForm):
    def __init__(self,
                 entity_model: EntityModel,
                 ledger_model: Union[str, UUID, LedgerModel],
                 *args, **kwargs
                 ):
        super().__init__(*args, **kwargs)
        self.ENTITY_MODEL: EntityModel = entity_model

        # processes the provided ledger model UUID is valid....
        if isinstance(ledger_model, (str, UUID)):
            ledger_model = self.ENTITY_MODEL.get_ledgers().get(uuid__exact=ledger_model)
        elif isinstance(ledger_model, LedgerModel):
            self.ENTITY_MODEL.validate_ledger_model_for_entity(ledger_model)

        self.LEDGER_MODEL: LedgerModel = ledger_model

        if 'timestamp' in self.fields:
            self.fields['timestamp'].required = False
        if 'entity_unit' in self.fields:
            self.fields['entity_unit'].queryset = self.ENTITY_MODEL.entityunitmodel_set.all()

    def clean(self):
        if self.LEDGER_MODEL.is_locked():
            raise ValidationError(message=_('Cannot create new Journal Entries on a locked Ledger.'))
        self.instance.ledger = self.LEDGER_MODEL
        return super().clean()

    class Meta:
        model = JournalEntryModel
        fields = [
            'timestamp',
            'entity_unit',
            'description'
        ]
        widgets = {
            'parent': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'entity_unit': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'timestamp': DateTimeInput(attrs={
                'type': 'datetime-local',
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'description': Textarea(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            })
        }
        labels = {
            'entity_unit': _('Entity Unit')
        }


class JournalEntryModelUpdateForm(ModelForm):

    def clean_timestamp(self):
        if 'timestamp' in self.changed_data:
            new_timestamp = self.cleaned_data['timestamp']
            je_model: JournalEntryModel = self.instance
            if je_model.is_in_locked_period(new_timestamp=new_timestamp):
                raise ValidationError(
                    message=_(f'Invalid timestamp {self.cleaned_data["timestamp"]} due to Closing Entries.')
                )
        return self.cleaned_data['timestamp']

    class Meta(JournalEntryModelCreateForm.Meta):
        model = JournalEntryModel
        fields = [
            'timestamp',
            'entity_unit',
            'description'
        ]


class JournalEntryModelCannotEditForm(JournalEntryModelUpdateForm):
    class Meta(JournalEntryModelCreateForm.Meta):
        fields = [
            'description'
        ]



from django.forms import ModelForm, TextInput, ValidationError
from django.utils.translation import gettext_lazy as _

from django_ledger.models import EntityUnitModel
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class EntityUnitModelCreateForm(ModelForm):
    def __init__(self, *args, entity_slug, user_model, **kwargs):
        super().__init__(*args, **kwargs)
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model

    def clean_name(self):
        name = self.cleaned_data['name']
        if len(name) < 10:
            raise ValidationError(_('Unit name must be at least 10 characters long'))
        return name

    class Meta:
        model = EntityUnitModel
        fields = [
            'name',
            'document_prefix'
        ]
        widgets = {
            'name': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'document_prefix': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
        }


class EntityUnitModelUpdateForm(EntityUnitModelCreateForm):
    pass



from behave import *


def parse_float(value):
    return float(value)


register_type(FloatNumber=parse_float)


@given("that we know what the two numbers are {n1:FloatNumber} and {n2:FloatNumber}")
def step_impl(context, n1: float, n2: float):
    context.n1 = n1
    context.n2 = n2


@when("we add both numbers")
def step_impl(context):
    assert isinstance(context.n1, float)
    assert isinstance(context.n2, float)
    context.result = context.n1 + context.n2


@then("we get back the correct result {result:FloatNumber}")
def step_impl(context, result):
    assert context.result == result



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

A Vendor refers to the person or entity that provides products and services to the business for a fee.
Vendors are an integral part of the billing process as they are the providers of goods and services for the
business.

Vendors can be flagged as active/inactive or hidden. Vendors who no longer conduct business with the EntityModel,
whether temporarily or indefinitely may be flagged as inactive (i.e. active is False). Hidden Vendors will not show up
as an option in the UI, but can still be used programmatically (via API).
"""
import os
import warnings
from uuid import uuid4, UUID

from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db import models, transaction, IntegrityError
from django.db.models import Q, F, QuerySet, Manager
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _

from django_ledger.models.deprecations import deprecated_entity_slug_behavior
from django_ledger.models.mixins import ContactInfoMixIn, CreateUpdateMixIn, FinancialAccountInfoMixin, TaxInfoMixIn
from django_ledger.models.utils import lazy_loader
from django_ledger.settings import (
    DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING,
    DJANGO_LEDGER_VENDOR_NUMBER_PREFIX,
    DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR
)


def vendor_picture_upload_to(instance, filename):
    if not instance.customer_number:
        instance.generate_customer_number(commit=False)
    vendor_number = instance.customer_number
    name, ext = os.path.splitext(filename)
    safe_name = slugify(name)
    return f'vendor_pictures/{vendor_number}/{safe_name}{ext.lower()}'


class VendorModelValidationError(ValidationError):
    pass


class VendorModelQuerySet(QuerySet):
    """
    Custom defined VendorModel QuerySet.
    """

    def for_user(self, user_model) -> 'VendorModelQuerySet':
        if user_model.is_superuser:
            return self
        return self.filter(
            Q(entity_model__admin=user_model) |
            Q(entity_model__managers__in=[user_model])
        )

    def active(self) -> 'VendorModelQuerySet':
        """
        Active vendors can be assigned to new bills and show on dropdown menus and views.

        Returns
        -------
        VendorModelQuerySet
            A QuerySet of active Vendors.
        """
        return self.filter(active=True)

    def inactive(self) -> 'VendorModelQuerySet':
        """
        Active vendors can be assigned to new bills and show on dropdown menus and views.
        Marking VendorModels as inactive can help reduce Database load to populate select inputs and also inactivate
        VendorModels that are not relevant to the Entity anymore. Also, it makes de UI cleaner by not populating
        unnecessary choices.

        Returns
        -------
        VendorModelQuerySet
            A QuerySet of inactive Vendors.
        """
        return self.filter(active=False)

    def hidden(self) -> 'VendorModelQuerySet':
        """
        Hidden vendors do not show on dropdown menus, but may be used via APIs or any other method that does not
        involve the UI.

        Returns
        -------
        VendorModelQuerySet
            A QuerySet of hidden Vendors.
        """
        return self.filter(hidden=True)

    def visible(self) -> 'VendorModelQuerySet':
        """
        Visible vendors show on dropdown menus and views. Visible vendors are active and not hidden.

        Returns
        -------
        VendorModelQuerySet
            A QuerySet of visible Vendors.
        """
        return self.filter(
            Q(hidden=False) & Q(active=True)
        )


class VendorModelManager(Manager):
    """
    Manages operations related to VendorModel instances.

    A specialized manager for handling interactions with VendorModel entities,
    providing additional support for filtering based on associated EntityModel or EntityModel slug.
    """

    @deprecated_entity_slug_behavior
    def for_entity(self, entity_model: 'EntityModel | str | UUID' = None, **kwargs) -> VendorModelQuerySet:
        """
            Filters the queryset for a given entity model.

            This method modifies the queryset to include only those records
            associated with the specified entity model. The entity model can
            be provided in various formats such as an instance of `EntityModel`,
            a string representing the entity's slug, or its UUID.

            If a deprecated parameter `user_model` is provided, it will issue
            a warning and may alter the behavior if the deprecated behavior flag
            `DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR` is set.

            Parameters
            ----------
            entity_model : EntityModel | str | UUID
                The entity model or its identifier (slug or UUID) to filter the
                queryset by.

            **kwargs
                Additional parameters for optional functionality. The parameter
                `user_model` is supported for backward compatibility but is
                deprecated and should be avoided.

            Returns
            -------
            VendorModelQuerySet
                A queryset filtered for the given entity model.
        """
        EntityModel = lazy_loader.get_entity_model()

        qs = self.get_queryset()
        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )

            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(entity_model=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(entity_model__slug__exact=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(entity_model_id=entity_model)
        else:
            raise VendorModelValidationError(
                'EntityModel slug must be either a string or an EntityModel instance'
            )
        return qs


class VendorModelAbstract(ContactInfoMixIn,
                          FinancialAccountInfoMixin,
                          TaxInfoMixIn,
                          CreateUpdateMixIn):
    """
    This is the main abstract class which the VendorModel database will inherit from.
    The VendorModel inherits functionality from the following MixIns:

        1. :func:`ContactInfoMixIn <django_ledger.models.mixins.ContactInfoMixIn>`
        2. :func:`BankAccountInfoMixIn <django_ledger.models.mixins.BankAccountInfoMixIn>`
        3. :func:`TaxInfoMixIn <django_ledger.models.mixins.TaxInfoMixIn>`
        4. :func:`CreateUpdateMixIn <django_ledger.models.mixins.CreateUpdateMixIn>`

    Attributes
    __________

    uuid : UUID
        This is a unique primary key generated for the table. The default value of this field is uuid4().

    entity_model: EntityModel
        The EntityModel associated with this Vendor.

    vendor_name: str
        A string representing the name the customer uses to do business with the EntityModel.

    vendor_number: str
        A unique, auto-generated human-readable number which identifies the vendor within the EntityModel.

    description: str
        A text field to capture the description about the vendor.

    active: bool
        We can set any vendor to be active or inactive. Defaults to True.

    hidden: bool
        Hidden VendorModel don't show on the UI. Defaults to False.

    additional_info: dict
        Any additional information about the vendor, stored as a JSON object using a JSONField.


    """
    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    vendor_code = models.SlugField(
        max_length=50,
        null=True,
        blank=True,
        verbose_name='User defined vendor code.'
    )
    vendor_number = models.CharField(max_length=30,
                                     null=True,
                                     blank=True,
                                     editable=False,
                                     verbose_name=_('Vendor Number'), help_text='System generated vendor number.')
    vendor_name = models.CharField(max_length=100)

    entity_model = models.ForeignKey('django_ledger.EntityModel',
                                     on_delete=models.CASCADE,
                                     verbose_name=_('Vendor Entity'),
                                     editable=False)
    description = models.TextField()
    active = models.BooleanField(default=True)
    hidden = models.BooleanField(default=False)
    picture = models.ImageField(upload_to=vendor_picture_upload_to, null=True, blank=True)

    additional_info = models.JSONField(null=True, blank=True, default=dict)

    objects = VendorModelManager.from_queryset(queryset_class=VendorModelQuerySet)()

    class Meta:
        verbose_name = _('Vendor')
        indexes = [
            models.Index(fields=['entity_model', 'vendor_number']),
            models.Index(fields=['vendor_number']),
            models.Index(fields=['created']),
            models.Index(fields=['updated']),
            models.Index(fields=['active']),
            models.Index(fields=['hidden']),
        ]
        unique_together = [
            ('entity_model', 'vendor_number')
        ]
        abstract = True

    def __str__(self):
        if not self.vendor_number:
            f'Unknown Vendor: {self.vendor_name}'
        return f'{self.vendor_number}: {self.vendor_name}'

    def can_generate_vendor_number(self) -> bool:
        """
        Determines if the VendorModel can be issued a Vendor Number.
        VendorModel have a unique sequential number, which is unique for each EntityModel/VendorModel.

        Returns
        -------
        bool
            True if the vendor number can be generated, else False.
        """
        return all([
            self.entity_model_id,
            not self.vendor_number
        ])

    def _get_next_state_model(self, raise_exception: bool = True):
        """
        Fetches the updated EntityStateModel associated with the vendor number sequence.
        If EntityStateModel is not present, a new model will be created.

        Parameters
        ----------
        raise_exception: bool
            Raises IntegrityError if Database cannot determine the next EntityStateModel available.

        Returns
        -------
        EntityStateModel
            The EntityStateModel associated with the VendorModel number sequence.
        """
        EntityStateModel = lazy_loader.get_entity_state_model()

        try:
            LOOKUP = {
                'entity_model_id__exact': self.entity_model_id,
                'key__exact': EntityStateModel.KEY_VENDOR
            }

            state_model_qs = EntityStateModel.objects.filter(**LOOKUP).select_for_update()
            state_model = state_model_qs.get()
            state_model.sequence = F('sequence') + 1
            state_model.save()
            state_model.refresh_from_db()

            return state_model
        except ObjectDoesNotExist:

            LOOKUP = {
                'entity_model_id': self.entity_model_id,
                'entity_unit_id': None,
                'fiscal_year': None,
                'key': EntityStateModel.KEY_VENDOR,
                'sequence': 1
            }
            state_model = EntityStateModel.objects.create(**LOOKUP)
            return state_model
        except IntegrityError as e:
            if raise_exception:
                raise e

    def generate_vendor_number(self, commit: bool = False) -> str:
        """
        Atomic Transaction. Generates the next Vendor Number available.

        Parameters
        __________

        commit: bool
            Commits transaction into VendorModel. Defaults to False.

        Returns
        _______
        str
            A String, representing the current VendorModel instance document number.
        """
        if self.can_generate_vendor_number():
            with transaction.atomic(durable=True):

                state_model = None
                while not state_model:
                    state_model = self._get_next_state_model(raise_exception=False)

            seq = str(state_model.sequence).zfill(DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING)
            self.vendor_number = f'{DJANGO_LEDGER_VENDOR_NUMBER_PREFIX}-{seq}'

            if commit:
                self.save(update_fields=['vendor_number', 'updated'])

        return self.vendor_number

    def clean(self):
        """
        Custom defined clean method that fetches the next vendor number if not yet fetched.
        Additional validation may be provided.
        """
        if self.can_generate_vendor_number():
            self.generate_vendor_number(commit=False)

    def save(self, **kwargs):
        """
        Custom-defined save method that automatically fetches the vendor number if not present.

        Parameters
        ----------
        kwargs
            Keywords passed to the super().save() method of the VendorModel.
        """
        if self.can_generate_vendor_number():
            self.generate_vendor_number(commit=False)
        super(VendorModelAbstract, self).save(**kwargs)


class VendorModel(VendorModelAbstract):
    """
    Base Vendor Model Implementation
    """

    class Meta(VendorModelAbstract.Meta):
        abstract = False



# Generated by Django 5.1.5 on 2025-01-17 00:11

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('django_ledger', '0018_transactionmodel_cleared_transactionmodel_reconciled_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='transactionmodel',
            name='amount',
            field=models.DecimalField(decimal_places=2, default=0.0, help_text='Amount of the transaction.',
                                      max_digits=20, validators=[django.core.validators.MinValueValidator(0)],
                                      verbose_name='Amount'),
        ),
        migrations.AlterField(
            model_name='transactionmodel',
            name='description',
            field=models.CharField(blank=True,
                                   help_text='A description to be included with this individual transaction.',
                                   max_length=100, null=True, verbose_name='Transaction Description'),
        ),
        migrations.AlterField(
            model_name='transactionmodel',
            name='tx_type',
            field=models.CharField(choices=[('credit', 'Credit'), ('debit', 'Debit')], max_length=10,
                                   verbose_name='Transaction Type'),
        ),
    ]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

A Bank Account refers to the financial institution which holds financial assets for the EntityModel.
A bank account usually holds cash, which is a Current Asset. Transactions may be imported using the open financial
format specification OFX into a staging area for final disposition into the EntityModel ledger.
"""
import warnings
from typing import Optional
from uuid import uuid4, UUID

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q, QuerySet, Manager
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _

from django_ledger.models import CreateUpdateMixIn, FinancialAccountInfoMixin
from django_ledger.models.deprecations import deprecated_entity_slug_behavior
from django_ledger.models.utils import lazy_loader
from django_ledger.settings import DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR

UserModel = get_user_model()


class BankAccountValidationError(ValidationError):
    pass


class BankAccountModelQuerySet(QuerySet):
    """
    A custom-defined QuerySet for the BankAccountModel.
    """

    def for_user(self, user_model) -> 'BankAccountModelQuerySet':
        if user_model.is_superuser:
            return self
        return self.filter(
            Q(entity_model__admin=user_model) |
            Q(entity_model__managers__in=[user_model])
        )

    def active(self) -> 'BankAccountModelQuerySet':
        """
        Active bank accounts which can be used to create new transactions.

        Returns
        _______
        BankAccountModelQuerySet
            A filtered BankAccountModelQuerySet of active accounts.
        """
        return self.filter(active=True)

    def hidden(self) -> 'BankAccountModelQuerySet':
        """
        Hidden bank accounts which can be used to create new transactions. but will not show in drop down menus
        in the UI.

        Returns
        _______
        BankAccountModelQuerySet
            A filtered BankAccountModelQuerySet of active accounts.
        """
        return self.filter(hidden=True)


class BankAccountModelManager(Manager):
    """
    Custom defined Model Manager for the BankAccountModel.
    """

    def get_queryset(self) -> BankAccountModelQuerySet:
        return BankAccountModelQuerySet(self.model, using=self._db)

    @deprecated_entity_slug_behavior
    def for_entity(self, entity_model: 'EntityModel | str | UUID' = None, **kwargs) -> BankAccountModelQuerySet:
        """
        Allows only the authorized user to query the BankAccountModel for a given EntityModel.
        This is the recommended initial QuerySet.

        Parameters
        __________
        entity_slug: str or EntityModel
            The entity slug or EntityModel used for filtering the QuerySet.
        """
        EntityModel = lazy_loader.get_entity_model()

        qs = self.get_queryset()
        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(entity_model=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(entity_model__slug__exact=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(entity_model_id=entity_model)
        else:
            raise BankAccountValidationError(
                message=_('Must pass EntityModel slug or EntityModel UUID'),
            )
        return qs


class BankAccountModelAbstract(FinancialAccountInfoMixin, CreateUpdateMixIn):
    """
    This is the main abstract class which the BankAccountModel database will inherit from.
    The BankAccountModel inherits functionality from the following MixIns:

        1. :func:`BankAccountInfoMixIn <django_ledger.models.mixins.BankAccountInfoMixIn>`
        2. :func:`CreateUpdateMixIn <django_ledger.models.mixins.CreateUpdateMixIn>`


    Attributes
    ----------
    uuid : UUID
        This is a unique primary key generated for the table. The default value of this field is uuid4().
    name: str
        A user defined name for the bank account as a String.
    entity_model: EntityModel
        The EntityModel associated with the BankAccountModel instance.
    account_model: AccountModel
        The AccountModel associated with the BankAccountModel instance. Must be an account with role ASSET_CA_CASH.
    active: bool
        Determines whether the BackAccountModel instance bank account is active. Defaults to True.
    hidden: bool
        Determines whether the BackAccountModel instance bank account is hidden. Defaults to False.
    """

    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)

    # todo: rename to account_name?...
    name = models.CharField(max_length=150, null=True, blank=True)
    entity_model = models.ForeignKey('django_ledger.EntityModel',
                                     on_delete=models.CASCADE,
                                     verbose_name=_('Entity Model'))

    account_model = models.ForeignKey('django_ledger.AccountModel',
                                      on_delete=models.RESTRICT,
                                      help_text=_(
                                          'Account model be used to map transactions from financial institution'
                                      ),
                                      verbose_name=_('Associated Account Model'))
    active = models.BooleanField(default=False)
    hidden = models.BooleanField(default=False)
    objects = BankAccountModelManager()

    def configure(self,
                  entity_slug,
                  user_model: Optional[UserModel],
                  commit: bool = False):

        EntityModel = lazy_loader.get_entity_model()
        if isinstance(entity_slug, str):
            if not user_model:
                raise BankAccountValidationError(_('Must pass user_model when using entity_slug.'))
            entity_model_qs = EntityModel.objects.for_user(user_model=user_model)
            entity_model = get_object_or_404(entity_model_qs, slug__exact=entity_slug)
        elif isinstance(entity_slug, EntityModel):
            entity_model = entity_slug
        else:
            raise BankAccountValidationError('entity_slug must be an instance of str or EntityModel')

        self.entity_model = entity_model
        self.clean()
        if commit:
            self.save(update_fields=[
                'entity_model',
                'updated'
            ])
        return self, entity_model

    def is_active(self):
        return self.active is True

    class Meta:
        abstract = True
        verbose_name = _('Bank Account')
        indexes = [
            models.Index(fields=['account_type']),
            models.Index(fields=['account_model']),
            models.Index(fields=['entity_model'])
        ]
        unique_together = [
            ('entity_model', 'account_number'),
            ('entity_model', 'account_model', 'account_number', 'routing_number')
        ]

    def __str__(self):
        return f'{self.get_account_type_display()} Bank Account: {self.name}'

    def can_hide(self) -> bool:
        return self.hidden is False

    def can_unhide(self) -> bool:
        return self.hidden is True

    def can_activate(self) -> bool:
        return self.active is False

    def can_inactivate(self) -> bool:
        return self.active is True

    def mark_as_active(self, commit: bool = False, raise_exception: bool = True, **kwargs):
        if not self.can_activate():
            if raise_exception:
                raise BankAccountValidationError('Bank Account cannot be activated.')
        self.active = True
        if commit:
            self.save(update_fields=[
                'active',
                'updated'
            ])

    def mark_as_inactive(self, commit: bool = False, raise_exception: bool = True, **kwargs):
        if not self.can_inactivate():
            if raise_exception:
                raise BankAccountValidationError('Bank Account cannot be deactivated.')
        self.active = False
        if commit:
            self.save(update_fields=[
                'active',
                'updated'
            ])


class BankAccountModel(BankAccountModelAbstract):
    """
    Base Bank Account Model Implementation
    """

    class Meta(BankAccountModelAbstract.Meta):
        abstract = False



# Generated by Django 5.0.4 on 2024-04-25 13:41

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0015_remove_chartofaccountmodel_locked_and_more'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='accountmodel',
            name='django_ledg_coa_mod_e19964_idx',
        ),
        migrations.AlterField(
            model_name='ledgermodel',
            name='entity',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_ledger.entitymodel', verbose_name='Ledger Entity'),
        ),
        migrations.AlterField(
            model_name='ledgermodel',
            name='ledger_xid',
            field=models.SlugField(allow_unicode=True, blank=True, help_text='User Defined Ledger ID', max_length=150, null=True, verbose_name='Ledger External ID'),
        ),
        migrations.AddIndex(
            model_name='accountmodel',
            index=models.Index(fields=['coa_model', 'code'], name='django_ledg_coa_mod_e073bc_idx'),
        ),
        migrations.AddIndex(
            model_name='accountmodel',
            index=models.Index(fields=['code'], name='django_ledg_code_081adc_idx'),
        ),
        migrations.AddIndex(
            model_name='ledgermodel',
            index=models.Index(fields=['entity', 'ledger_xid'], name='django_ledg_entity__7be095_idx'),
        ),
        migrations.AddIndex(
            model_name='ledgermodel',
            index=models.Index(fields=['ledger_xid'], name='django_ledg_ledger__05f099_idx'),
        ),
    ]
